"""The `mesh_struct.py` module provides classes and methods for handling and manipulating mesh structures in the
context of neuroimaging and brain stimulation studies. It includes classes for handling tetrahedral meshes and
regions of interest (ROIs).

The module includes the following classes:

- `TetrahedraLinear`: This class represents a mesh consisting of linear tetrahedra. It provides methods for
calculating various quantities of interest (QOIs) in the mesh, interpolating data between nodes and elements,
calculating the electric field and current density, and more.

- `Mesh`: This class is a general mesh class to initialize default attributes and provides methods to fill default
values based on the approach, write the mesh data to an HDF5 file, and print the mesh information.

- `ROI`: This class represents a region of interest (ROI) in the mesh. It provides methods to write the ROI data to
an HDF5 file and print the ROI information.

Each class and method in this module is documented with docstrings providing more detailed information about its
purpose, parameters, and return values.

This module is primarily used for handling and visualizing data related to neuroimaging and brain stimulation studies.
"""
import os
import json
import time
import nibabel
import numpy as np
from numpy import cross as cycross
import pynibs

__package__ = "pynibs"


def __path__():
    return os.path.dirname(__file__)


class TetrahedraLinear:
    """
    Mesh, consisting of linear tetrahedra.

    Parameters
    ----------
    points : array of float [N_points x 3]
        Vertices of FE mesh
    triangles : np.ndarray of int [N_tri x 3]
        Connectivity of points forming triangles
    triangles_regions : np.ndarray of int [N_tri x 1]
        Region identifiers of triangles
    tetrahedra : np.ndarray of int [N_tet x 4]
        Connectivity of points forming tetrahedra
    tetrahedra_regions : np.ndarray of int [N_tet x 1]
        Region identifiers of tetrahedra

    Attributes
    ----------
    N_points : int
        Number of vertices
    N_tet : int
        Number of tetrahedra
    N_tri : int
        Number of triangles
    N_region : int
        Number of regions
    region : np.ndarray of int
        Region labels
    tetrahedra_volume : np.ndarray of float [N_tet x 1]
        Volumes of tetrahedra
    tetrahedra_center : np.ndarray of float [N_tet x 1]
        Center of tetrahedra
    triangles_center : np.ndarray of float [N_tri x 1]
        Center of triangles
    triangles_normal : np.ndarray of float [N_tri x 3]
        Normal components of triangles pointing outwards
    """

    def __init__(self, points, triangles, triangles_regions, tetrahedra, tetrahedra_regions):
        """ Initialize TetrahedraLinear class """
        self.points = points
        self.triangles = triangles
        self.triangles_regions = triangles_regions
        self.tetrahedra = tetrahedra
        self.tetrahedra_regions = tetrahedra_regions
        # index of points in "tetrahedra" start with 0 or 1

        self.tetrahedra_triangle_surface_idx = - np.ones((self.triangles.shape[0], 2))

        # shift index to start always from 0 (python)
        if self.tetrahedra.size != 0:
            self.idx_start = np.min(self.tetrahedra)
            self.tetrahedra = self.tetrahedra - self.idx_start
            self.N_tet = self.tetrahedra.shape[0]
            p1_tet = self.points[self.tetrahedra[:, 0], :]  # [P1x P1y P1z]
            p2_tet = self.points[self.tetrahedra[:, 1], :]
            p3_tet = self.points[self.tetrahedra[:, 2], :]
            p4_tet = self.points[self.tetrahedra[:, 3], :]
            self.tetrahedra_volume = pynibs.calc_tetrahedra_volume_cross(p1_tet, p2_tet, p3_tet, p4_tet)
            self.tetrahedra_center = 1.0 / 4 * (p1_tet + p2_tet + p3_tet + p4_tet)

        else:
            self.N_tet = 0
            self.idx_start = 0
        self.triangles = self.triangles - self.idx_start

        self.region = np.unique(self.tetrahedra_regions)

        # number of elements and points etc
        self.N_points = self.points.shape[0]
        self.N_tri = self.triangles.shape[0]
        self.N_region = len(self.region)

        # count index lists of elements [0,1,2,....]
        self.tetrahedra_index = np.arange(self.N_tet)
        self.triangles_index = np.arange(self.N_tri)

        if self.N_tri > 0:
            p1_tri = self.points[self.triangles[:, 0], :]
            p2_tri = self.points[self.triangles[:, 1], :]
            p3_tri = self.points[self.triangles[:, 2], :]

            self.triangles_center = 1.0 / 3 * (p1_tri + p2_tri + p3_tri)
            self.triangles_normal = cycross(p2_tri - p1_tri, p3_tri - p1_tri)
            normal_norm = np.linalg.norm(self.triangles_normal, axis=1)
            normal_norm = normal_norm[:, np.newaxis]
            self.triangles_normal = self.triangles_normal / np.tile(normal_norm, (1, 3))
            self.triangles_area = 0.5 * np.linalg.norm(np.cross(p2_tri - p1_tri, p3_tri - p1_tri), axis=1)

    def calc_E_on_GM_WM_surface_simnibs(self, phi, dAdt, roi, subject, verbose=False, mesh_idx=0):
        """
        Determines the normal and tangential component of the induced electric field on a GM-WM surface by recalculating
        phi and dA/dt in an epsilon environment around the GM/WM surface (upper and lower GM-WM surface) or by using
        the Simnibs interpolation function.

        Parameters
        ----------
        phi : np.ndarray of float
            (N_nodes, 1) Scalar electric potential given in the nodes of the mesh.
        dAdt : np.ndarray of float
            (N_nodes, 3) Magnetic vector potential given in the nodes of the mesh.
        roi : pynibs.mesh.mesh_struct.ROI
            RegionOfInterestSurface object class instance.
        subject : pynibs.subject.Subject
            Subject object loaded from .hdf5 file.
        verbose : bool
            Print information to stdout.
        mesh_idx : int
            Mesh index.

        Returns
        -------
        E_normal : np.ndarray of float
            (N_points, 3) Normal vector of electric field on GM-WM surface.
        E_tangential : np.ndarray of float
            (N_points, 3) Tangential vector of electric field on GM-WM surface.
        """
        import tempfile
        import simnibs.msh.mesh_io as mesh_io
        import simnibs.simulation.fem as fem
        import simnibs.msh.transformations as transformations

        mesh_folder = subject.mesh[mesh_idx]["mesh_folder"]

        # load mesh
        mesh = mesh_io.read_msh(subject.mesh[mesh_idx]["fn_mesh_msh"])

        # write phi and dAdt in msh
        dAdt_SimNIBS = mesh_io.NodeData(dAdt, name='D', mesh=mesh)
        phi_SimNIBS = mesh_io.NodeData(phi.flatten(), name='v', mesh=mesh)

        if verbose:
            print("Calculating e-field")
        out = fem.calc_fields(phi_SimNIBS, "vDEe", cond=None, dadt=dAdt_SimNIBS)

        with tempfile.TemporaryDirectory() as f:
            fn_res_tmp = os.path.join(f, "res.msh")
            # mesh_io.write_msh(out, fn_res_tmp)

            if verbose:
                print("Interpolating values to midlayer of GM")
            # determine e in midlayer
            transformations.middle_gm_interpolation(mesh_fn=out,
                                                    m2m_folder=os.path.join(mesh_folder, "m2m_" + subject.id),
                                                    out_folder=f,
                                                    out_fsaverage=None,
                                                    depth=0.5,
                                                    quantities=['norm', 'normal', 'tangent', 'angle'],
                                                    fields=None,
                                                    open_in_gmsh=False,
                                                    write_msh=False)  #

            # load freesurfer surface
            if type(roi.gm_surf_fname) is not list:
                roi.gm_surf_fname = [roi.gm_surf_fname]

            points_gm = [None for _ in range(len(roi.gm_surf_fname))]
            con_gm = [None for _ in range(len(roi.gm_surf_fname))]

            max_idx_gm = 0

            if (roi.gm_surf_fname is list and len(roi.gm_surf_fname) > 0) or (roi.gm_surf_fname is str):
                fn_surface = list(roi.gm_surf_fname)
            elif (roi.midlayer_surf_fname is list and len(roi.gm_surf_fname) > 0) or (roi.midlayer_surf_fname is str):
                fn_surface = list(roi.midlayer_surf_fname)

            for i in range(len(fn_surface)):
                points_gm[i], con_gm[i] = nibabel.freesurfer.read_geometry(os.path.join(mesh_folder, fn_surface[i]))

                con_gm[i] = con_gm[i] + max_idx_gm

                max_idx_gm = max_idx_gm + points_gm[i].shape[0]  # np.max(con_gm[i]) + 2

            points_gm = np.vstack(points_gm)
            con_gm = np.vstack(con_gm)

            if verbose:
                print("Processing data to ROI")
            if roi.fn_mask is None or roi.fn_mask == []:

                if roi.X_ROI is None or roi.X_ROI == []:
                    roi.X_ROI = [-np.inf, np.inf]
                if roi.Y_ROI is None or roi.Y_ROI == []:
                    roi.Y_ROI = [-np.inf, np.inf]
                if roi.Z_ROI is None or roi.Z_ROI == []:
                    roi.Z_ROI = [-np.inf, np.inf]

                roi_mask_bool = (roi.node_coord_mid[:, 0] > min(roi.X_ROI)) & (
                        roi.node_coord_mid[:, 0] < max(roi.X_ROI)) & \
                                (roi.node_coord_mid[:, 1] > min(roi.Y_ROI)) & (
                                        roi.node_coord_mid[:, 1] < max(roi.Y_ROI)) & \
                                (roi.node_coord_mid[:, 2] > min(roi.Z_ROI)) & (
                                        roi.node_coord_mid[:, 2] < max(roi.Z_ROI))
                roi_mask_idx = np.where(roi_mask_bool)

            else:
                if type(roi.fn_mask) is np.ndarray:
                    if roi.fn_mask.ndim == 0:
                        roi.fn_mask = roi.fn_mask.astype(str).tolist()

                # read mask from freesurfer mask file
                mask = nibabel.freesurfer.mghformat.MGHImage.from_filename(
                    os.path.join(mesh_folder, roi.fn_mask)).dataobj[:]
                roi_mask_idx = np.where(mask > 0.5)

            # read results data
            if verbose:
                print("Reading SimNIBS midlayer data")
            e_normal = []
            e_tan = []

            for fn_surf in fn_surface:
                if "lh" in os.path.split(fn_surf)[1]:
                    e_normal.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "lh.res.central.E." + "normal")).flatten()[:, np.newaxis])
                    e_tan.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "lh.res.central.E." + "tangent")).flatten()[:, np.newaxis])

                if "rh" in os.path.split(fn_surf)[1]:
                    e_normal.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "rh.res.central.E." + "normal")).flatten()[:, np.newaxis])
                    e_tan.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "rh.res.central.E." + "tangent")).flatten()[:, np.newaxis])

            e_normal = np.vstack(e_normal)
            e_tan = np.vstack(e_tan)

            # transform point data to element data
            if verbose:
                print("Transforming point data to element data")
            e_normal = pynibs.data_nodes2elements(data=e_normal, con=con_gm)
            e_tan = pynibs.data_nodes2elements(data=e_tan, con=con_gm)

            # crop results data to ROI
            # if not roi_mask_bool.all():
            if roi_mask_idx:
                if verbose:
                    print("Cropping results data to ROI")

                # get row index where all points are lying inside ROI
                con_row_idx = [i for i in range(con_gm.shape[0]) if len(np.intersect1d(con_gm[i,], roi_mask_idx)) == 3]

                e_normal = e_normal[con_row_idx, :]
                e_tan = e_tan[con_row_idx, :]

        return e_normal, e_tan

    def calc_E_on_GM_WM_surface_simnibs_KW(self, phi, dAdt, roi, subject, verbose=False, mesh_idx=0):
        """
        Determines the normal and tangential component of the induced electric field on a GM-WM surface by recalculating
        phi and dA/dt in an epsilon environment around the GM/WM surface (upper and lower GM-WM surface) or by using
        the Simnibs interpolation function.

        Parameters
        ----------
        phi : np.ndarray of float
            (N_nodes, 1) Scalar electric potential given in the nodes of the mesh.
        dAdt : np.ndarray of float
            (N_nodes, 1) Magnetic vector potential given in the nodes of the mesh.
        roi : pynibs.mesh.mesh_struct.ROI
            RegionOfInterestSurface object class instance.
        subject : pynibs.subject.Subject
            Subject object loaded from .hdf5 file.
        verbose : bool
            Print information to stdout.
        mesh_idx : int
            Mesh index.

        Returns
        -------
        E_normal : np.ndarray of float
            (N_points, 3) Normal vector of electric field on GM-WM surface.
        E_tangential : np.ndarray of float
            (N_points, 3) Tangential vector of electric field on GM-WM surface.
        """
        import tempfile
        import simnibs.msh.mesh_io as mesh_io
        import simnibs.simulation.fem as fem
        import simnibs.msh.transformations as transformations

        mesh_folder = subject.mesh[mesh_idx]["mesh_folder"]

        # load mesh
        mesh = mesh_io.read_msh(subject.mesh[mesh_idx]["fn_mesh_msh"])

        # write phi and dAdt in msh
        dadt_simnibs = mesh_io.NodeData(dAdt, name='D', mesh=mesh)
        phi_simnibs = mesh_io.NodeData(phi.flatten(), name='v', mesh=mesh)

        if verbose:
            print("Calculating e-field")
        out = fem.calc_fields(phi_simnibs, "vDEe", cond=None, dadt=dadt_simnibs)

        with tempfile.TemporaryDirectory() as f:
            fn_res_tmp = os.path.join(f, "res.msh")
            mesh_io.write_msh(out, fn_res_tmp)

            if verbose:
                print("Interpolating values to midlayer of GM")
            # determine e in midlayer
            transformations.middle_gm_interpolation(mesh_fn=fn_res_tmp,
                                                    m2m_folder=os.path.join(mesh_folder, "m2m_" + subject.id),
                                                    out_folder=f,
                                                    out_fsaverage=None,
                                                    depth=0.5,
                                                    quantities=['norm', 'normal', 'tangent', 'angle'],
                                                    fields=None,
                                                    open_in_gmsh=False)  # write_msh=False

            # load freesurfer surface
            if type(roi.gm_surf_fname) is not list:
                roi.gm_surf_fname = [roi.gm_surf_fname]

            points_gm = [None for _ in range(len(roi.gm_surf_fname))]
            con_gm = [None for _ in range(len(roi.gm_surf_fname))]

            max_idx_gm = 0

            if (type(roi.gm_surf_fname) is list and roi.gm_surf_fname[0] is not None) or \
                    (type(roi.gm_surf_fname) is str):
                if type(roi.gm_surf_fname) is str:
                    fn_surface = [roi.gm_surf_fname]
                else:
                    fn_surface = roi.gm_surf_fname

            elif (type(roi.midlayer_surf_fname) is list and roi.gm_surf_fname is not None) or \
                    (type(roi.midlayer_surf_fname) is str):
                if type(roi.midlayer_surf_fname) is str:
                    fn_surface = [roi.midlayer_surf_fname]
                else:
                    fn_surface = roi.midlayer_surf_fname

            for i in range(len(fn_surface)):
                points_gm[i], con_gm[i] = nibabel.freesurfer.read_geometry(os.path.join(mesh_folder, fn_surface[i]))

                con_gm[i] = con_gm[i] + max_idx_gm

                max_idx_gm = max_idx_gm + points_gm[i].shape[0]  # np.max(con_gm[i]) + 2

            points_gm = np.vstack(points_gm)
            con_gm = np.vstack(con_gm)

            if verbose:
                print("Processing data to ROI")
            if roi.fn_mask is None or roi.fn_mask == []:

                if roi.X_ROI is None or roi.X_ROI == []:
                    roi.X_ROI = [-np.inf, np.inf]
                if roi.Y_ROI is None or roi.Y_ROI == []:
                    roi.Y_ROI = [-np.inf, np.inf]
                if roi.Z_ROI is None or roi.Z_ROI == []:
                    roi.Z_ROI = [-np.inf, np.inf]

                roi_mask_bool = (roi.node_coord_mid[:, 0] > min(roi.X_ROI)) & (
                        roi.node_coord_mid[:, 0] < max(roi.X_ROI)) & \
                                (roi.node_coord_mid[:, 1] > min(roi.Y_ROI)) & (
                                        roi.node_coord_mid[:, 1] < max(roi.Y_ROI)) & \
                                (roi.node_coord_mid[:, 2] > min(roi.Z_ROI)) & (
                                        roi.node_coord_mid[:, 2] < max(roi.Z_ROI))
                roi_mask_idx = np.where(roi_mask_bool)

            else:
                if type(roi.fn_mask) is np.ndarray:
                    if roi.fn_mask.ndim == 0:
                        roi.fn_mask = roi.fn_mask.astype(str).tolist()

                # read mask from freesurfer mask file
                mask = nibabel.freesurfer.mghformat.MGHImage.from_filename(
                    os.path.join(mesh_folder, roi.fn_mask)).dataobj[:]
                roi_mask_idx = np.where(mask > 0.5)

            # read results data
            if verbose:
                print("Reading SimNIBS midlayer data")
            e_normal = []
            e_tan = []

            for fn_surf in fn_surface:
                if "lh" in os.path.split(fn_surf)[1]:
                    e_normal.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "lh.res.central.E." + "normal")).flatten()[:, np.newaxis])
                    e_tan.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "lh.res.central.E." + "tangent")).flatten()[:, np.newaxis])

                if "rh" in os.path.split(fn_surf)[1]:
                    e_normal.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "rh.res.central.E." + "normal")).flatten()[:, np.newaxis])
                    e_tan.append(nibabel.freesurfer.read_morph_data(
                        os.path.join(f, "rh.res.central.E." + "tangent")).flatten()[:, np.newaxis])

            e_normal = np.vstack(e_normal)
            e_tan = np.vstack(e_tan)

            # transform point data to element data
            if verbose:
                print("Transforming point data to element data")
            e_normal = pynibs.data_nodes2elements(data=e_normal, con=con_gm)
            e_tan = pynibs.data_nodes2elements(data=e_tan, con=con_gm)

            # crop results data to ROI
            # if not roi_mask_bool.all():
            if roi_mask_idx:
                if verbose:
                    print("Cropping results data to ROI")

                # get row index where all points are lying inside ROI
                con_row_idx = [i for i in range(con_gm.shape[0]) if len(np.intersect1d(con_gm[i,], roi_mask_idx)) == 3]

                e_normal = e_normal[con_row_idx, :]
                e_tan = e_tan[con_row_idx, :]

        return e_normal, e_tan

    def calc_E_on_GM_WM_surface3(self, phi, dAdt, roi, verbose=True, mode="components"):
        """
        Determines the normal and tangential component of the induced electric field on a GM-WM surface by recalculating
        phi and dA/dt in an epsilon environment around the GM/WM surface (upper and lower GM-WM surface).

        Parameters
        ----------
        phi : np.ndarray of float
            (N_nodes, 1) Scalar electric potential given in the nodes of the mesh.
        dAdt : np.ndarray of float
            (N_nodes, 3) Magnetic vector potential given in the nodes of the mesh.
        roi : pynibs.mesh.mesh_struct.ORI
            RegionOfInterestSurface object class instance.
        verbose : bool
            Print information to stdout.
        mode : str
            Select mode of output:
            - "components" : return x, y, and z component of tangential and normal components
            - "magnitude" : return magnitude of tangential and normal component (normal with sign for direction)

        Returns
        -------
        E_normal : np.ndarray of float
            (N_nodes, 3) Normal vector of electric field on GM-WM surface.
        E_tangential : np.ndarray of float
            (N_nodes, 3) Tangential vector of electric field on GM-WM surface.
        """
        # check if dimension are fitting
        assert phi.shape[0] == dAdt.shape[0]
        assert dAdt.shape[1] == 3

        # interpolate electric scalar potential to central points of upper and lower surface triangles
        if verbose:
            print("Interpolating electric scalar potential to central points of upper and lower surface triangles")
        phi_gm_wm_surface_up = self.calc_QOI_in_points_tet_idx(qoi=phi,
                                                               points_out=roi.tri_center_coord_up,
                                                               tet_idx=roi.tet_idx_tri_center_up.flatten())

        phi_gm_wm_surface_low = self.calc_QOI_in_points_tet_idx(qoi=phi,
                                                                points_out=roi.tri_center_coord_low,
                                                                tet_idx=roi.tet_idx_tri_center_low.flatten())

        # determine distance between upper and lower surface (in m!)
        d = np.linalg.norm(roi.tri_center_coord_up - roi.tri_center_coord_low, axis=1)[:, np.newaxis] * 1E-3
        d[np.argwhere(d == 0)[:, 0]] = 1e-6  # delete zero distances

        # determine surface normal vector (normalized)
        # n = ((points_up - points_low) / np.tile(d, (1, 3)))*1E-3
        # n = (points_up - points_low) * 1E-3

        p1_tri = roi.node_coord_mid[roi.node_number_list[:, 0], :]
        p2_tri = roi.node_coord_mid[roi.node_number_list[:, 1], :]
        p3_tri = roi.node_coord_mid[roi.node_number_list[:, 2], :]

        n = cycross(p2_tri - p1_tri, p3_tri - p1_tri)
        normal_norm = np.linalg.norm(n, axis=1)
        normal_norm = normal_norm[:, np.newaxis]
        n = n / np.tile(normal_norm, (1, 3))

        # interpolate magnetic vector potential to central surface points (primary electric field)
        # E_pri = griddata(self.points, dAdt, surf_mid, method='linear', fill_value=np.NaN, rescale=False)
        if verbose:
            print("Interpolating magnetic vector potential to central surface points (primary electric field)")
        e_pri = self.calc_QOI_in_points_tet_idx(qoi=dAdt,
                                                points_out=roi.tri_center_coord_mid,
                                                tet_idx=roi.tet_idx_tri_center_mid.flatten())

        # determine its normal component
        e_pri_normal = np.multiply(np.sum(np.multiply(e_pri, n), axis=1)[:, np.newaxis], n)

        # determine gradient of phi and multiply with surface normal (secondary electric field)
        e_sec_normal = np.multiply((phi_gm_wm_surface_up - phi_gm_wm_surface_low) * 1E-3 / d, n)

        # combine (normal) primary and secondary electric field
        e_normal = self.calc_E(e_sec_normal, e_pri_normal)

        # compute tangential component of secondary electric field on surface
        if verbose:
            print("Interpolating scalar electric potential to nodes of midlayer (primary electric field)")
        phi_surf_mid_nodes = self.calc_QOI_in_points_tet_idx(qoi=phi,
                                                             points_out=roi.node_coord_mid,
                                                             tet_idx=roi.tet_idx_node_coord_mid.flatten())

        if verbose:
            print("Determine gradient of scalar electric potential on midlayer surface (E_sec_tangential)")
        e_sec_tan = pynibs.calc_gradient_surface(phi=phi_surf_mid_nodes,
                                                 points=roi.node_coord_mid,
                                                 triangles=roi.node_number_list)

        # compute tangential component of primary electric field on surface
        e_pri_tan = e_pri - e_pri_normal

        # compute tangential component of total electric field
        e_tan = self.calc_E(e_sec_tan, e_pri_tan)

        # determine total E on surface (sanity check)
        # E = self.calc_QOI_in_points(E, surf_mid)

        if mode == "magnitude":
            # get sign info of normal component
            e_normal_dir = (np.sum(e_normal * n, axis=1) > 0)[:, np.newaxis].astype(int)

            e_normal_dir[e_normal_dir == 1] = 1
            e_normal_dir[e_normal_dir == 0] = -1

            # determine magnitude of vectors and assign sign info
            e_tan = np.linalg.norm(e_tan, axis=1)[:, np.newaxis]
            e_normal = np.linalg.norm(e_normal, axis=1)[:, np.newaxis] * e_normal_dir

        return e_normal, e_tan

    def calc_E_on_GM_WM_surface(self, E, roi):
        """
        Determines the normal and tangential component of the induced electric field on a GM-WM surface using
        nearest neighbour principle.

        Parameters
        ----------
        E : np.ndarray of float [N_tri x 3]
            Induced electric field given in the tetrahedra centre of the mesh instance
        roi : pynibs.roi.RegionOfInterestSurface
            RegionOfInterestSurface object class instance

        Returns
        -------
        E_normal : np.ndarray of float [N_points x 3]
            Normal vector of electric field on GM-WM surface
        E_tangential : np.ndarray of float [N_points x 3]
            Tangential vector of electric field on GM-WM surface
        """

        e_gm_wm_surface = E[roi.tet_idx_nodes_mid, :]

        # determine surface normal vector (normalized)
        n = cycross(roi.node_coord_mid[roi.node_number_list[:, 1]] - roi.node_coord_mid[roi.node_number_list[:, 0]],
                    roi.node_coord_mid[roi.node_number_list[:, 2]] - roi.node_coord_mid[roi.node_number_list[:, 0]])
        n = n / np.linalg.norm(n, axis=1)[:, np.newaxis]

        # determine its normal component
        e_normal = np.multiply(np.sum(np.multiply(e_gm_wm_surface, n), axis=1)[:, np.newaxis], n)

        # compute tangential component of total electric field
        e_tan = e_gm_wm_surface - e_normal

        # determine total E on surface (sanity check)
        # E = self.calc_QOI_in_points(E, surf_mid)

        return e_normal, e_tan

    def calc_QOI_in_points(self, qoi, points_out):
        """
        Calculate QOI_out in points_out using the mesh instance and the quantity of interest (QOI).

        Parameters
        ----------
        qoi : np.ndarray of float
            Quantity of interest in nodes of tetrahedra mesh instance
        points_out : np.ndarray of float
            Point coordinates (x, y, z) where the qoi is going to be interpolated by linear basis functions

        Returns
        -------
        qoi_out : np.ndarray of float
            Quantity of interest in points_out

        """

        n_phi_points_out = points_out.shape[0]
        qoi_out = np.zeros(
            [n_phi_points_out, qoi.shape[1] if qoi.ndim > 1 else 1])

        p1_all = self.points[self.tetrahedra[:, 0], :]
        p2_all = self.points[self.tetrahedra[:, 1], :]
        p3_all = self.points[self.tetrahedra[:, 2], :]
        p4_all = self.points[self.tetrahedra[:, 3], :]

        # identify  in which tetrahedron the point lies
        # (all other volumes have at least one negative sub-volume)

        # determine all volumes (replacing points with points_out)
        # find the element where all volumes are > 0 (not inverted element)
        # get index of this tetrahedron
        # do it successively to decrease amount of volume calculations for all
        # 4 points in tetrahedra
        for i in range(n_phi_points_out):
            start = time.time()
            vtest1 = pynibs.calc_tetrahedra_volume_cross(np.tile(points_out[i, :], (p1_all.shape[0], 1)),
                                                         p2_all,
                                                         p3_all,
                                                         p4_all)
            tet_idx_bool_1 = (vtest1 >= 0)
            tet_idx_1 = np.nonzero(tet_idx_bool_1)[0]

            vtest2 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx_1, :],
                                                         np.tile(
                                                             points_out[i, :], (tet_idx_1.shape[0], 1)),
                                                         p3_all[tet_idx_1, :],
                                                         p4_all[tet_idx_1, :])
            tet_idx_bool_2 = (vtest2 >= 0)
            tet_idx_2 = tet_idx_1[np.nonzero(tet_idx_bool_2)[0]]

            vtest3 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx_2, :],
                                                         p2_all[tet_idx_2, :],
                                                         np.tile(
                                                             points_out[i, :], (tet_idx_2.shape[0], 1)),
                                                         p4_all[tet_idx_2, :])
            tet_idx_bool_3 = (vtest3 >= 0)
            tet_idx_3 = tet_idx_2[np.nonzero(tet_idx_bool_3)[0]]

            vtest4 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx_3, :],
                                                         p2_all[tet_idx_3, :],
                                                         p3_all[tet_idx_3, :],
                                                         np.tile(points_out[i, :], (tet_idx_3.shape[0], 1)))
            tet_idx_bool_4 = (vtest4 >= 0)
            tet_idx = tet_idx_3[np.nonzero(tet_idx_bool_4)[0]]

            # calculate subvolumes of final tetrahedron and its total volume
            vsub1 = pynibs.calc_tetrahedra_volume_cross(points_out[i, :][np.newaxis],
                                                        p2_all[tet_idx, :],
                                                        p3_all[tet_idx, :],
                                                        p4_all[tet_idx, :])
            vsub2 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                        points_out[i, :][np.newaxis],
                                                        p3_all[tet_idx, :],
                                                        p4_all[tet_idx, :])
            vsub3 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                        p2_all[tet_idx, :],
                                                        points_out[i, :][np.newaxis],
                                                        p4_all[tet_idx, :])
            vsub4 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                        p2_all[tet_idx, :],
                                                        p3_all[tet_idx, :],
                                                        points_out[i, :][np.newaxis], )

            vsub = np.array([vsub1, vsub2, vsub3, vsub4])
            vtot = np.sum(vsub)

            # calculate phi_out
            qoi_out[i,] = 1.0 * np.dot(vsub.T, qoi[self.tetrahedra[tet_idx[0], :],]) / vtot

            stop = time.time()
            print(('Total: Point: {:d}/{:d} [{} sec]\n'.format(i + 1, n_phi_points_out, stop - start)))

        return qoi_out

    def calc_QOI_in_points_tet_idx(self, qoi, points_out, tet_idx):
        """
        Calculate QOI_out in points_out sitting in tet_idx using the mesh instance and the quantity of interest (QOI).

        Parameters
        ----------
        qoi : np.ndarray of float
            Quantity of interest in nodes of tetrahedra mesh instance
        points_out : np.ndarray of float
            Point coordinates (x, y, z) where the qoi is going to be interpolated by linear basis functions
        tet_idx : np.ndarray of int
            Element indices where the points_out are sitting

        Returns
        -------
        qoi_out : np.ndarray of float
            Quantity of interest in points_out

        """

        n_phi_points_out = points_out.shape[0]
        qoi_out = np.zeros([n_phi_points_out, qoi.shape[1] if qoi.ndim > 1 else 1])

        p1_all = self.points[self.tetrahedra[:, 0], :]
        p2_all = self.points[self.tetrahedra[:, 1], :]
        p3_all = self.points[self.tetrahedra[:, 2], :]
        p4_all = self.points[self.tetrahedra[:, 3], :]

        # determine sub-volumes
        vsub1 = pynibs.calc_tetrahedra_volume_cross(points_out,
                                                    p2_all[tet_idx, :],
                                                    p3_all[tet_idx, :],
                                                    p4_all[tet_idx, :])
        vsub2 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                    points_out,
                                                    p3_all[tet_idx, :],
                                                    p4_all[tet_idx, :])
        vsub3 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                    p2_all[tet_idx, :],
                                                    points_out,
                                                    p4_all[tet_idx, :])
        vsub4 = pynibs.calc_tetrahedra_volume_cross(p1_all[tet_idx, :],
                                                    p2_all[tet_idx, :],
                                                    p3_all[tet_idx, :],
                                                    points_out)
        vsub = np.hstack([vsub1, vsub2, vsub3, vsub4])
        vtot = np.sum(vsub, axis=1)

        # calculate the QOIs in the tetrahedron of interest
        for i in range(qoi.shape[1]):
            qoi_out[:, i] = 1.0 * np.sum(np.multiply(vsub, qoi[self.tetrahedra[tet_idx, :], i]), axis=1) / vtot

        # for i in range(N_phi_points_out):
        #     # calculate subvolumes of final tetrahedron and its total volume
        #     Vsub1 = pynibs.calc_tetrahedra_volume_cross(points_out[i, :][np.newaxis],
        #                                               p2_all[tet_idx[i], :][np.newaxis],
        #                                               p3_all[tet_idx[i], :][np.newaxis],
        #                                               P4_all[tet_idx[i], :][np.newaxis])
        #     vsub2 = pynibs.calc_tetrahedra_volume_cross(P1_all[tet_idx[i], :][np.newaxis],
        #                                               points_out[i, :][np.newaxis],
        #                                               p3_all[tet_idx[i], :][np.newaxis],
        #                                               P4_all[tet_idx[i], :][np.newaxis])
        #     Vsub3 = pynibs.calc_tetrahedra_volume_cross(P1_all[tet_idx[i], :][np.newaxis],
        #                                               p2_all[tet_idx[i], :][np.newaxis],
        #                                               points_out[i, :][np.newaxis],
        #                                               P4_all[tet_idx[i], :][np.newaxis])
        #     Vsub4 = pynibs.calc_tetrahedra_volume_cross(P1_all[tet_idx[i], :][np.newaxis],
        #                                               p2_all[tet_idx[i], :][np.newaxis],
        #                                               p3_all[tet_idx[i], :][np.newaxis],
        #                                               points_out[i, :][np.newaxis])
        #
        #     vtot = np.sum([Vsub1, vsub2, Vsub3, Vsub4])
        #
        #     # calculate the QOIs in the tetrahedron of interest
        #     qoi_out[i,] = 1.0 * np.dot(vsub.T, qoi[self.tetrahedra[tet_idx[i], :],]) / vtot

        return qoi_out

    def data_nodes2elements(self, data):
        """
        Interpolate data given in the nodes to the tetrahedra center.

        Parameters
        ----------
        data : np.ndarray [N_nodes x N_data]
            Data in nodes

        Returns
        -------
        data_elements : np.ndarray [N_elements x N_data]
            Data in elements
        """
        data_elements = np.sum(data[self.tetrahedra[:, i]] for i in range(4)) / 4.0

        return data_elements

    def data_elements2nodes(self, data):
        """
        Transforms an data in tetrahedra into the nodes after Zienkiewicz et al. (1992) [1]_.
        Can only transform volume data, i.e. needs the data in the surrounding tetrahedra to average it to the nodes.
        Will not work well for discontinuous fields (like E, if several tissues are used).

        Parameters
        ----------
        data : np.ndarray [N_elements x N_data]
            Data in tetrahedra

        Returns
        -------
        data_nodes : np.ndarray [N_nodes x N_data]
            Data in nodes

        Notes
        -----
        .. [1] Zienkiewicz, Olgierd Cecil, and Jian Zhong Zhu. "The superconvergent patch recovery and a
           posteriori error estimates. Part 1: The recovery technique." International Journal for
           Numerical Methods in Engineering 33.7 (1992): 1331-1364.
        """

        # check dimension of input data
        if data.ndim == 1:
            data = data[:, np.newaxis]

        n_data = data.shape[1]
        data_nodes = np.zeros((self.N_points, n_data))

        if self.N_tet != data.shape[0]:
            raise ValueError("The number of data points in the data has to be equal to the number"
                             "of elements in the mesh")

        value = np.atleast_2d(data)
        if value.shape[0] < value.shape[1]:
            value = value.T

        # nd = np.zeros((self.N_points, N_data))

        # get all nodes used in tetrahedra, creates the NodeData structure
        # uq = np.unique(msh.elm[msh.elm.tetrahedra])
        # nd = NodeData(np.zeros((len(uq), self.nr_comp)), self.field_name, mesh=msh)
        # nd.node_number = uq

        # Get the point in the outside surface
        points_outside = np.unique(self.get_outside_faces())
        outside_points_mask = np.in1d(self.tetrahedra, points_outside).reshape(-1, 4)
        masked_th_nodes = np.copy(self.tetrahedra)
        masked_th_nodes[outside_points_mask] = -1

        # Calculates the quantities needed for the superconvergent patch recovery
        uq_in, th_nodes = np.unique(masked_th_nodes, return_inverse=True)

        baricenters = self.tetrahedra_center
        volumes = self.tetrahedra_volume
        baricenters = np.hstack([np.ones((baricenters.shape[0], 1)), baricenters])

        A = np.empty((len(uq_in), 4, 4))
        b = np.empty((len(uq_in), 4, n_data), 'float64')
        for i in range(4):
            for j in range(i, 4):
                A[:, i, j] = np.bincount(th_nodes.reshape(-1),
                                         np.repeat(baricenters[:, i], 4) *
                                         np.repeat(baricenters[:, j], 4))
        A[:, 1, 0] = A[:, 0, 1]
        A[:, 2, 0] = A[:, 0, 2]
        A[:, 3, 0] = A[:, 0, 3]
        A[:, 2, 1] = A[:, 1, 2]
        A[:, 3, 1] = A[:, 1, 3]
        A[:, 3, 2] = A[:, 2, 3]

        for j in range(n_data):
            for i in range(4):
                b[:, i, j] = np.bincount(th_nodes.reshape(-1),
                                         np.repeat(baricenters[:, i], 4) *
                                         np.repeat(value[:, j], 4))

        a = np.linalg.solve(A[1:], b[1:])
        p = np.hstack([np.ones((len(uq_in) - 1, 1)), self.points[uq_in[1:]]])
        f = np.einsum('ij, ijk -> ik', p, a)
        data_nodes[uq_in[1:]] = f

        # Assigns the average value to the points in the outside surface
        masked_th_nodes = np.copy(self.tetrahedra)
        masked_th_nodes[~outside_points_mask] = -1
        uq_out, th_nodes_out = np.unique(masked_th_nodes, return_inverse=True)

        sum_vals = np.empty((len(uq_out), n_data), 'float64')

        for j in range(n_data):
            sum_vals[:, j] = np.bincount(th_nodes_out.reshape(-1),
                                         np.repeat(value[:, j], 4) *
                                         np.repeat(volumes, 4))

        sum_vols = np.bincount(th_nodes_out.reshape(-1), np.repeat(volumes, 4))

        data_nodes[uq_out[1:]] = (sum_vals / sum_vols[:, None])[1:]

        return data_nodes

    def get_outside_faces(self, tetrahedra_indices=None):
        """
        Creates a list of nodes in each face that are in the outer volume.

        Parameters
        ----------
        tetrahedra_indices : np.ndarray
            Indices of the tetrehedra where the outer volume is to be determined (default: all tetrahedra)

        Returns
        -------
        faces : np.ndarray
            List of nodes in faces in arbitrary order
        """

        if tetrahedra_indices is None:
            tetrahedra_indices = self.tetrahedra_index

        th = self.tetrahedra[tetrahedra_indices]
        faces = th[:, [[0, 2, 1], [0, 1, 3], [0, 3, 2], [1, 2, 3]]]
        faces = faces.reshape(-1, 3)
        hash_array = np.array([hash(f.tobytes()) for f in np.sort(faces, axis=1)])
        unique, idx, inv, count = np.unique(hash_array, return_index=True,
                                            return_inverse=True, return_counts=True)

        # if np.any(count > 2):
        #     raise ValueError('Invalid Mesh: Found a face with more than 2 adjacent'
        #                      ' tetrahedra!')

        outside_faces = faces[idx[count == 1]]

        return outside_faces

    def calc_gradient(self, phi):
        """
        Calculate gradient of scalar DOF in tetrahedra center.

        Parameters
        ----------
        phi : np.ndarray of float [N_nodes]
            Scalar DOF the gradient is calculated for

        Returns
        -------
        grad_phi : np.ndarray of float [N_tet x 3]
            Gradient of Scalar DOF in tetrahedra center
        """

        a1 = np.vstack((self.points[self.tetrahedra[:, 3], :] - self.points[self.tetrahedra[:, 1], :],
                        self.points[self.tetrahedra[:, 2], :] -
                        self.points[self.tetrahedra[:, 0], :],
                        self.points[self.tetrahedra[:, 3], :] -
                        self.points[self.tetrahedra[:, 0], :],
                        self.points[self.tetrahedra[:, 1], :] - self.points[self.tetrahedra[:, 0], :]))

        a2 = np.vstack((self.points[self.tetrahedra[:, 2], :] - self.points[self.tetrahedra[:, 1], :],
                        self.points[self.tetrahedra[:, 3], :] -
                        self.points[self.tetrahedra[:, 0], :],
                        self.points[self.tetrahedra[:, 1], :] -
                        self.points[self.tetrahedra[:, 0], :],
                        self.points[self.tetrahedra[:, 2], :] - self.points[self.tetrahedra[:, 0], :]))

        # a3 = np.vstack((self.points[self.tetrahedra[:, 0], :] - self.points[self.tetrahedra[:, 1], :],
        #                 self.points[self.tetrahedra[:, 1], :] -
        #                 self.points[self.tetrahedra[:, 0], :],
        #                 self.points[self.tetrahedra[:, 2], :] -
        #                 self.points[self.tetrahedra[:, 0], :],
        #                 self.points[self.tetrahedra[:, 3], :] - self.points[self.tetrahedra[:, 0], :]))

        volumes = np.sum(np.multiply(cycross(a1, a2), a3), 1)
        volumes = volumes[:, np.newaxis]
        dlambda = np.transpose(np.reshape(cycross(
            a1, a2) / np.tile(volumes, (1, 3)), (self.N_tet, 4, 3), order='F'), (0, 2, 1))

        grad_phi = np.zeros((self.N_tet, 3))
        # calculate gradient at barycenters of tetrahedra
        for j in range(4):
            grad_phi = grad_phi + dlambda[:, :, j] * np.tile(phi[self.tetrahedra[:, j]], (1, 3))

        return grad_phi

    def calc_E(self, grad_phi, omegaA):
        """
        Calculate electric field with gradient of electric potential and omega-scaled magnetic vector potential A.

        .. math:: \mathbf{E}=-\\nabla\\varphi-\omega\mathbf{A}

        Parameters
        ----------
        grad_phi : np.ndarray of float
            (N_tet, 3) Gradient of Scalar DOF in tetrahedra center.
        omegaA : np.ndarray of float
            (N_tet, 3) Magnetic vector potential in tetrahedra center (scaled with angular frequency omega).

        Returns
        -------
        E : np.ndarray of float
            (N_tet, 3) Electric field in tetrahedra center.
        """
        e = -grad_phi - omegaA
        return e

    def calc_J(self, E, sigma):
        """
        Calculate current density J. The conductivity sigma is a list of np.arrays containing conductivities of
        regions (scalar and/or tensor).

        .. math::
            \mathbf{J} = [\sigma]\mathbf{E}

        Parameters
        ----------
        E : np.ndarray of float
            (N_tet, 3) Electric field in tetrahedra center.
        sigma : list of np.ndarray of float
            [N_regions](3, 3) Conductivities of regions (scalar and/or tensor).

        Returns
        -------
        E : np.ndarray of float
            (N_tet, 3) Electric field in tetrahedra center.
        """
        j = np.zeros((E.shape[0], 3))

        for i in range(self.N_region):
            tet_bool_idx = self.tetrahedra_regions == self.region[i]
            j[tet_bool_idx[:, 0], :] = np.dot(
                sigma[i], E[tet_bool_idx[:, 0], :].T).T
        return j

    def calc_surface_adjacent_tetrahedra_idx_list(self, fname):
        """
        Determine the indices of the tetrahedra touching the surfaces and save the indices into a .txt file specified
        with fname.

        Parameters
        ----------
        fname : str
            Filename of output .txt file.

        Returns
        -------
        <File> : .txt file
            Element indices of the tetrahedra touching the surfaces (outer-most elements)
        """
        # determine indices of the 2 adjacent tetrahedra with common face on
        # surface
        # P1_idx = np.zeros((self.N_tet, 1), dtype=bool)
        # p2_idx = np.zeros((self.N_tet, 1), dtype=bool)
        # p3_idx = np.zeros((self.N_tet, 1), dtype=bool)
        tet_idx_pos = np.zeros((self.N_tri, 1)).astype(int)
        tet_idx_neg = np.zeros((self.N_tri, 1)).astype(int)

        start = time.time()

        tetrahedra0 = self.tetrahedra[:, 0]
        tetrahedra1 = self.tetrahedra[:, 1]
        tetrahedra2 = self.tetrahedra[:, 2]
        tetrahedra3 = self.tetrahedra[:, 3]

        for i in range(self.N_tri):

            if not (i % 100) and i > 0:
                stop = time.time()
                print(('Tri: {:d}/{:d} [{} sec]\n'.format(i, self.N_tri, stop - start)))
                start = time.time()

            triangle = set(self.triangles[i, :])

            triangle0 = self.triangles[i, 0]
            triangle1 = self.triangles[i, 1]
            triangle2 = self.triangles[i, 2]

            p1_idx = (tetrahedra0 == triangle0) | (tetrahedra1 == triangle0) | (
                    tetrahedra2 == triangle0) | (tetrahedra3 == triangle0)
            p2_idx = (tetrahedra0 == triangle1) | (tetrahedra1 == triangle1) | (
                    tetrahedra2 == triangle1) | (tetrahedra3 == triangle1)
            p3_idx = (tetrahedra0 == triangle2) | (tetrahedra1 == triangle2) | (
                    tetrahedra2 == triangle2) | (tetrahedra3 == triangle2)

            tet_bool_idx = p1_idx & p2_idx & p3_idx
            tet_idx = np.where(tet_bool_idx)[0][:]

            # get 4th (test) point of e.g. first tetrahedron which is not in
            # plane
            p4_idx = list(set(self.tetrahedra[tet_idx[0], :]) - triangle)

            # calculate projection of the line between:
            # center of triangle -> 4th point
            # and
            # normal of the triangle
            c = np.dot(
                self.points[p4_idx, :] - self.triangles_center[i, :], self.triangles_normal[i, :])

            # positive projection: normal points to the 4th (test) point of first tetrahedron
            # and first tetrahedron is on "positive" side

            # outermost surface (has only one adjacent tetrahedron)
            if len(tet_idx) == 1:
                if c > 0:
                    tet_idx_pos[i] = tet_idx[0]
                    tet_idx_neg[i] = -1

                else:
                    tet_idx_pos[i] = -1
                    tet_idx_neg[i] = tet_idx[0]

            # inner surfaces have 2 adjacent tetrahedra
            else:
                if c > 0:
                    tet_idx_pos[i] = tet_idx[0]
                    tet_idx_neg[i] = tet_idx[1]
                else:
                    tet_idx_pos[i] = tet_idx[1]
                    tet_idx_neg[i] = tet_idx[0]

        # save the indices of the tetrahedra sharing the surfaces (negative,
        # i.e. bottom side first)
        self.tetrahedra_triangle_surface_idx = np.hstack(
            [tet_idx_neg, tet_idx_pos])
        f = open(fname, 'w')
        np.savetxt(f, self.tetrahedra_triangle_surface_idx, '%d')
        f.close()

    def calc_E_normal_tangential_surface(self, E, fname):
        """
        Calculate normal and tangential component of electric field on given surfaces of mesh instance.

        Parameters
        ----------
        E : np.ndarray of float [N_tri x 3]
            Electric field data on surfaces
        fname : str
            Filename of the .txt file containing the tetrahedra indices, which are adjacent to the surface triangles
            generated by the method "calc_surface_adjacent_tetrahedra_idx_list(self, fname)"

        Returns
        -------
        En_pos : np.ndarray of float [N_tri x 3]
            Normal component of electric field of top side (outside) of surface
        En_neg : np.ndarray of float [N_tri x 3]
            Normal component of electric field of bottom side (inside) of surface
        n : np.ndarray of float [N_tri x 3]
            Normal vector
        Et : np.ndarray of float [N_tri x 3]
            Tangential component of electric field lying in surface
        t : np.ndarray of float [N_tri x 3]
            Tangential vector
        """

        n = self.triangles_normal
        en_pos = np.zeros((self.N_tri, 1))
        en_neg = np.zeros((self.N_tri, 1))
        et = np.zeros((self.N_tri, 1))
        t = np.zeros((self.N_tri, 3))
        self.tetrahedra_triangle_surface_idx = np.loadtxt(fname).astype(int)

        for i in range(self.N_tri):
            en_neg[i, 0] = np.dot(
                E[self.tetrahedra_triangle_surface_idx[i, 0], :], n[i, :])

            if self.tetrahedra_triangle_surface_idx[i, 1] > -1:
                en_pos[i, 0] = np.dot(
                    E[self.tetrahedra_triangle_surface_idx[i, 1], :], n[i, :])
            else:
                en_pos[i, 0] = np.nan

            t[i, :] = E[self.tetrahedra_triangle_surface_idx[i, 0], :] - \
                      1.0 * en_neg[i, 0] * n[i, :]
            et[i, 0] = np.linalg.norm(t[i, :])
            t[i, :] = t[i, :] / et[i, 0] if et[i, 0] > 0 else np.zeros(3)

        return en_pos, en_neg, n, et, t

    def get_faces(self, tetrahedra_indexes=None):
        """
        Creates a list of nodes in each face and a list of faces in each tetrahedra.

        Parameters
        ----------
        tetrahedra_indexes : np.ndarray
            Indices of the tetrehedra where the faces are to be determined (default: all tetrahedra)

        Returns
        -------
        faces : np.ndarray
            List of nodes in faces, in arbitrary order
        th_faces : np.ndarray
            List of faces in each tetrahedra, starts at 0, order=((0, 2, 1), (0, 1, 3), (0, 3, 2), (1, 2, 3))
        face_adjacency_list : np.ndarray
            List of tetrahedron adjacent to each face, filled with -1 if a face is in a
            single tetrahedron. Not in the normal element ordering, but only in the order
            the tetrahedra are presented
        """

        if tetrahedra_indexes is None:
            tetrahedra_indexes = np.arange(self.tetrahedra.shape[0])
        # th = self[tetrahedra_indexes]
        th = self.tetrahedra[tetrahedra_indexes, :]
        faces = th[:, [[0, 2, 1], [0, 1, 3], [0, 3, 2], [1, 2, 3]]]
        faces = faces.reshape(-1, 3)
        hash_array = np.array([hash(f.tobytes()) for f in np.sort(faces, axis=1)])
        unique, idx, inv, count = np.unique(hash_array, return_index=True,
                                            return_inverse=True, return_counts=True)
        faces = faces[idx]
        face_adjacency_list = -np.ones((len(unique), 2), dtype=int)
        face_adjacency_list[:, 0] = idx // 4

        # if np.any(count > 2):
        #     raise ValueError('Invalid Mesh: Found a face with more than 2 adjacent'
        #                      ' tetrahedra!')

        # Remove the faces already seen from consideration
        # Second round in order to make adjacency list
        # create a new array with a mask in the elements already seen
        mask = unique[-1] + 1
        hash_array_masked = np.copy(hash_array)
        hash_array_masked[idx] = mask
        # make another array, where we delete the elements we have already seen
        hash_array_reduced = np.delete(hash_array, idx)
        # Finds where each element of the second array is in the first array
        # (https://stackoverflow.com/a/8251668)
        hash_array_masked_sort = hash_array_masked.argsort()
        hash_array_repeated_pos = hash_array_masked_sort[
            np.searchsorted(hash_array_masked[hash_array_masked_sort], hash_array_reduced)]
        # Now find the index of the face corresponding to each element in the
        # hash_array_reduced
        faces_repeated = np.searchsorted(unique, hash_array_reduced)
        # Finally, fill out the second column in the adjacency list
        face_adjacency_list[faces_repeated, 1] = hash_array_repeated_pos // 4

        return faces, inv.reshape(-1, 4), face_adjacency_list


class Mesh:
    """"
    Mesh class to initialize default attributes.
    """

    def __init__(self, mesh_name, subject_id, subject_folder):
        self.subject_id = subject_id
        self.subject_folder = subject_folder
        self.name = mesh_name
        self.info = None
        self.approach = None  # 'mri2mesh', 'headreco', 'charm'
        self.mri_idx = None

        # default parameters
        self.mesh_folder = os.path.join(subject_folder, 'mesh', mesh_name)
        self.fn_mesh_msh = os.path.join(self.mesh_folder, f"{subject_id}.msh")
        self.fn_mesh_hdf5 = os.path.join(self.mesh_folder, f"{subject_id}.hdf5")
        self.fn_tensor_vn = f"d2c_{subject_id}{os.sep}dti_results_T1space{os.sep}DTI_conf_tensor.nii.gz"
        self.fn_mri_conform = f"T1.nii.gz"
        self.fn_lh_midlayer = f"fs_{subject_id}{os.sep}surf{os.sep}lh.central"
        self.fn_rh_midlayer = f"fs_{subject_id}{os.sep}surf{os.sep}rh.central"
        self.vertex_density = 1.0  # headreco
        self.numvertices = 100000  # mri2mesh

        # refinement parameters
        self.center = None
        self.radius = None
        self.element_size = None
        self.refine_domains = None
        self.smooth_domains = None

        self.fn_lh_wm = None
        self.fn_rh_wm = None
        self.fn_lh_gm = None
        self.fn_rh_gm = None
        self.fn_lh_gm_curv = None
        self.fn_rh_gm_curv = None

        # charm meshes
        self.smooth_skin = None
        self.refinement_roi = None
        self.refinemement_element_size = None

    def fill_defaults(self, approach):
        """
        Initializes attributes for a headreco mesh.

        Parameters
        ----------
        approach: str
            'headreco'
            'mri2mesh'
            'charm'
        """
        self.approach = approach
        if approach == 'headreco':
            self.fn_mesh_msh = os.path.join(self.mesh_folder, f"{self.subject_id}.msh")
            self.fn_mesh_hdf5 = os.path.join(self.mesh_folder, f"{self.subject_id}.hdf5")
            self.fn_tensor_vn = f"d2c_{self.subject_id}{os.sep}dti_results_T1space{os.sep}DTI_conf_tensor.nii.gz"
            self.fn_mri_conform = f"{self.subject_id}_T1fs_conform.nii.gz"
            self.fn_lh_midlayer = f"fs_{self.subject_id}{os.sep}surf{os.sep}lh.central"
            self.fn_rh_midlayer = f"fs_{self.subject_id}{os.sep}surf{os.sep}rh.central"

        elif approach == 'mri2mesh':
            self.fn_mesh_msh = os.path.join(self.mesh_folder, f"{self.subject_id}.msh")
            self.fn_mesh_hdf5 = os.path.join(self.mesh_folder, f"{self.subject_id}.hdf5")
            self.fn_tensor_vn = f"d2c_{self.subject_id}{os.sep}dti_results_T1space{os.sep}DTI_conf_tensor.nii.gz"
            self.fn_mri_conform = f"{self.subject_id}_T1fs_conform.nii.gz"
            self.fn_lh_gm_curv = f"fs_{self.subject_id}{os.sep}surf{os.sep}lh.curv.pial"
            self.fn_rh_gm_curv = f"fs_{self.subject_id}{os.sep}surf{os.sep}rh.curv.pial"

        elif approach == 'charm':
            self.fn_mesh_msh = os.path.join(self.mesh_folder, f"{self.subject_id}.msh")
            self.fn_mesh_hdf5 = os.path.join(self.mesh_folder, f"{self.subject_id}.hdf5")
            self.fn_tensor_vn = f"d2c_{self.subject_id}{os.sep}dti_results_T1space{os.sep}DTI_conf_tensor.nii.gz"
            self.fn_mri_conform = f"{self.subject_id}_T1.nii.gz"
            self.fn_lh_midlayer = f"m2m_{self.subject_id}{os.sep}surfaces{os.sep}lh.central.gii"
            self.fn_rh_midlayer = f"m2m_{self.subject_id}{os.sep}surfaces{os.sep}rh.central.gii"
            self.use_fs = False

        else:
            raise NotImplementedError(f"Approach {approach} not implemented.")

    def write_to_hdf5(self, fn_hdf5, check_file_exist=False, verbose=False):
        """
        Write this mesh' attributes to .hdf5 file.

        Parameters
        ----------
        fn_hdf5 : str
        check_file_exist : bool
            Check if provided filenames exist, warn if not.
        verbose : bool
            Print self information
        """

        pynibs.write_dict_to_hdf5(fn_hdf5=fn_hdf5, data=self.__dict__, folder=f"mesh/{self.name}",
                                  check_file_exist=check_file_exist)
        if verbose:
            self.print()

    def print(self):
        """
        Print self information.
        """
        n_left, n_right = int(32 - np.floor((len(self.name) + 10) / 2)), int(32 - np.ceil((len(self.name) + 10) / 2))
        n_left, n_right = np.max(n_left, 0), np.max(n_right, 0)
        print("    " + "=" * n_left + f"  Mesh {self.name}:  " + "=" * n_right)
        print("\t" + json.dumps(self.__dict__, sort_keys=False, indent="\t", ))
        print("    " + "=" * 64 + "\n")


class ROI:
    """
    Region of interest class to initialize default attributes.
    """

    def __init__(self, subject_id, roi_name, mesh_name):
        self.subject_id = subject_id
        self.name = roi_name
        self.mesh_name = mesh_name
        self.type = None  # 'surface' or 'volume'
        self.info = None
        self.template = None  # None, 'MNI', 'fsaverage', 'subject'
        self.gm_surf_fname = None
        self.wm_surf_fname = None
        self.midlayer_surf_fname = None
        self.delta = 0.5
        self.refine = False
        self.X_ROI = None
        self.Y_ROI = None
        self.Z_ROI = None
        self.center = None
        self.radius = None
        self.layer = 3
        self.fn_mask = None
        self.fn_mask_avg = None
        self.hemisphere = None
        self.midlayer_surf_fname = None
        self.tri_center_coord_mid = None

    def write_to_hdf5(self, fn_hdf5, check_file_exist=False, verbose=False):
        """
        Write this mesh' attributes to .hdf5 file.

        Parameters
        ----------
        fn_hdf5 : str
        check_file_exist : bool
            Check if provided filenames exist, warn if not.
        verbose : bool
            Print self information
        """

        pynibs.write_dict_to_hdf5(fn_hdf5=fn_hdf5, data=self.__dict__, folder=f"roi/{self.mesh_name}/{self.name}",
                                  check_file_exist=check_file_exist)
        if verbose:
            self.print()

    def print(self):
        """
        Print self information.
        """
        n_left, n_right = int(32 - np.floor((len(self.name) + 10) / 2)), int(32 - np.ceil((len(self.name) + 10) / 2))
        n_left, n_right = np.max(n_left, 0), np.max(n_right, 0)
        print("    " + "=" * n_left + f"  ROI {self.name}:  " + "=" * n_right)
        print("\t" + json.dumps(self.__dict__, sort_keys=False, indent="\t", ))
        print("    " + "=" * 64 + "\n")



"""
Functions used in our perspective on e-field based TMS dosing[1]_

References
----------
.. [1] Numssen, O., Kuhnke, P., Weise, K., & Hartwigsen, G. (2024).
   Electric-field-based dosing for TMS. Imaging Neuroscience, 2, 1-12.
   DOI: 10.1162/imag_a_00106
"""
import os
import h5py
import numpy as np
import pynibs


def get_intensity_e(e1, e2, target1, target2, radius1, radius2, headmesh,
                    rmt=1, roi='midlayer_lh_rh', verbose=False):
    """
    Computes the stimulator intensity adjustment factor based on the electric field.

    Parameters
    ----------
    e1 : str
        .hdf5 e field with midlayer.
    e2 : str
        .hdf5 e field with midlayer.
    target1 : np.ndarray (3,)
        Coordinates of cortical site of MT.
    target2 : np.ndarray (3,)
        Coordinates of cortical target site.
    radius1 : float
        Electric field of field1 is averaged over elements inside this radius around target1.
    radius2 : float
        Electric field of field2 is averaged over elements inside this radius around target2.
    headmesh : str
        .hdf5 headmesh.
    rmt : float, default: 1
        Resting motor threshold to be corrected.
    roi : str, default: 'midlayer_lh_rh'
        Name of roi. Expected to sit in ``mesh['/data/midlayer/roi_surface/']``.
    verbose : bool, default: False
        Flag indicating verbosity.

    Returns
    -------
    rmt_e_corr : float
        Adjusted stimulation intensity for target2.
    """

    with h5py.File(headmesh, 'r') as f:
        tris = f[f'/roi_surface/{roi}/tri_center_coord_mid'][:]

    idx, e_avg_target, e_target, t_idx_sphere = [], [], [], []
    for field, target, radius in zip([e1, e2], [target1, target2], [radius1, radius2]):
        idx.append(np.argmin(np.linalg.norm(tris - target, axis=1)))
        t_idx_sphere.append(np.where(np.linalg.norm(tris - tris[idx[-1]], axis=1) < radius)[0])
        with h5py.File(field, 'r') as e:
            e_avg_target.append(np.mean(e[f'/data/midlayer/roi_surface/{roi}/E_mag'][t_idx_sphere[-1]]))
            e_target.append(e[f'/data/midlayer/roi_surface/{roi}/E_mag'][idx[-1]])

    # determine scaling factor
    e_fac_avg = e_avg_target[0] / e_avg_target[1]
    e_fac = e_target[0] / e_target[1]
    rmt_e_corr = rmt * e_fac_avg

    if verbose:
        print(f"Target1: {target1}->{tris[idx[0]]}. E: {e_target[0]:2.4f}, {len(t_idx_sphere[0])} elms")
        print(f"Target2: {target2}->{tris[idx[1]]}. E: {e_target[1]:2.4f}, {len(t_idx_sphere[0])} elms")
        print(f"Efield normalization factor: {e_fac_avg:2.4f} ({e_fac:2.4f} for single elm).")
        # print(f"Center: {target} { tris_center[t_idx, ]}.")
        print(f"Given intensity {rmt}% is normalized to {rmt * e_fac_avg:2.4f}%.")

    return rmt_e_corr


def get_intensity_e_old(mesh1, mesh2, target1, target2, radius1, radius2, rmt=1, verbose=False):
    """
    Computes the stimulator intensity adjustment factor based on the electric field.
    Something weird is going on here - check simnibs coordinates of midlayer before usage.

    Parameters
    ----------
    mesh1 : str or simnibs.msh.mesh_io.Msh
        Midlayer mesh containing results of the optimal coil position of MT in the midlayer
        (e.g.: .../subject_overlays/00001.hd_fixed_TMS_1-0001_MagVenture_MCF_B65_REF_highres.ccd_scalar_central.msh)
    mesh2 : str or simnibs.msh.mesh_io.Msh
        Midlayer mesh containing results of the optimal coil position of the target in the midlayer
        (e.g.: .../subject_overlays/00001.hd_fixed_TMS_1-0001_MagVenture_MCF_B65_REF_highres.ccd_scalar_central.msh)
    target1 : np.ndarray
        (3,) Coordinates of cortical site of MT.
    target2 : np.ndarray
        (3,) Coordinates of cortical target site.
    radius1 : float
        Electric field in target 1 is averaged over elements inside this radius.
    radius2 : float
        Electric field in target 2 is averaged over elements inside this radius.
    rmt : float, default: 1
        Resting motor threshold, which will be corrected.
    verbose : bool, default: False
        Flag indicating verbosity.

    Returns
    -------
    rmt_e_corr : float
        Adjusted stimulation intensity for target2.
    """
    from simnibs.msh.mesh_io import read_msh

    # load mesh1 (MT) if filename is provided
    if isinstance(mesh1, str):
        if os.path.splitext(mesh1)[1] == ".msh":
            mesh1 = read_msh(mesh1)
        elif os.path.splitext(mesh1)[1] == ".hdf5":
            mesh1 = pynibs.load_mesh_hdf5(mesh1)

    # load mesh2 (target) if filename is provided
    if isinstance(mesh2, str):
        if os.path.splitext(mesh2)[1] == ".msh":
            mesh2 = read_msh(mesh2)
        elif os.path.splitext(mesh2)[1] == ".hdf5":
            mesh2 = pynibs.load_mesh_hdf5(mesh2)

    # load electric fields in midlayer and average electric field around sphere in targets
    e_avg_target = []
    for mesh, target, radius in zip([mesh1, mesh2], [target1, target2], [radius1, radius2]):
        nodes = mesh.nodes.node_coord
        tris = mesh.elm.node_number_list[:, :-1] - 1
        tris_center = np.mean(nodes[tris,], axis=1)

        e_norm_nodes = None
        for nodedata in mesh.nodedata:
            if nodedata.field_name == "E_norm":
                e_norm_nodes = nodedata.value

        e_norm_tris = np.mean(e_norm_nodes[tris], axis=1)

        # project targets to midlayer
        t_idx = np.argmin(np.linalg.norm(tris_center - target, axis=1))

        # get indices of surrounding elements in some radius
        t_idx_sphere = np.where(np.linalg.norm(tris_center - tris_center[t_idx,], axis=1) < radius)[0]

        # average e-field in this area
        e_avg_target.append(np.mean(e_norm_tris[t_idx_sphere]))

        print(f"Center: {target} {tris_center[t_idx,]}.")

    # determine scaling factor
    e_fac = e_avg_target[0] / e_avg_target[1]
    rmt_e_corr = rmt * e_fac

    if verbose:
        print(f"Efield normalized factor is: {e_fac:2.4f}.")
        # print(f"Center: {target} { tris_center[t_idx, ]}.")
        print(f"Given stimulatior intensity {rmt}% is normalized to new intensity {rmt * e_fac:2.4f}%.")

    return rmt_e_corr


def get_intensity_stokes(mesh, target1, target2, spat_grad=3, rmt=0, scalp_tag=1005, roi=None, verbose=False):
    """
    Computes the stimulator intensity adjustment factor according to Stokes et al. 2005
    (doi:10.1152/jn.00067.2005).
    Adjustment is based on target-scalp distance differences:
    adj = (Dist2-Dist1)*spat_grad

    Parameters
    ----------
    mesh : str or simnibs.msh.mesh_io.Msh
        Mesh of the head model.
    target1 : np.ndarray
        (3,) Coordinates of cortical site of MT.
    target2 : np.ndarray
        (3,) Coordinates of cortical target site.
    spat_grad : float, default: 3
        Spatial gradient.
    rmt : float, default: 0
        Resting motor threshold, which will be corrected.
    scalp_tag: int, default: 1005
        Tag in the mesh where the scalp is to be set.
    roi: np.ndarray, optional
        (3,N) Array of nodes to project targets onto.
    verbose : bool, default: False
        Print verbosity information.

    Returns
    -------
    rmt_stokes : float
        Adjusted stimulation intensity for target2.
    """
    from simnibs.msh.mesh_io import read_msh
    from pynibs.mesh import project_on_scalp

    # load mesh if filename is provided
    if isinstance(mesh, str):
        if os.path.splitext(mesh)[1] == ".msh":
            mesh = read_msh(mesh)
        elif os.path.splitext(mesh)[1] == ".hdf5":
            mesh = pynibs.load_mesh_hdf5(mesh)

    t1_proj = project_on_scalp(target1, mesh, scalp_tag=scalp_tag)
    t2_proj = project_on_scalp(target2, mesh, scalp_tag=scalp_tag)

    if roi is not None:
        t1_idx = np.argmin(np.linalg.norm(roi - target1, axis=1))
        t2_idx = np.argmin(np.linalg.norm(roi - target2, axis=1))
        t1_on_roi = roi[t1_idx]
        t2_on_roi = roi[t2_idx]

        if verbose:
            print("Projecting targets on ROI:\n"
                  "T1: [{0:+06.2f}, {1:+06.2f}, {2:+06.2f}] -> [{3:+06.2f}, {4:+06.2f}, {5:+06.2f}] Dist: {6:05.2f}mm"
                  "\n".format(*target1, *t1_on_roi, np.linalg.norm(target1 - t1_on_roi)) + \
                  "T2: [{0:+06.2f}, {1:+06.2f}, {2:+06.2f}] -> [{3:+06.2f}, {4:+06.2f}, {5:+06.2f}] Dist: {6:05.2f}mm"
                  "".format(*target2, *t2_on_roi, np.linalg.norm(target2 - t2_on_roi)))
        target1 = t1_on_roi
        target2 = t2_on_roi

    t1_dist = np.linalg.norm(target1 - t1_proj)
    t2_dist = np.linalg.norm(target2 - t2_proj)

    stokes_factor = (t2_dist - t1_dist) * spat_grad
    rmt_stokes = rmt + stokes_factor

    if verbose:
        print("Target 1: [{0:+06.2f}, {1:+06.2f}, {2:+06.2f}] ->"
              " [{3:+06.2f}, {4:+06.2f}, {5:+06.2f}] Dist: {6:05.2f}mm ".format(*target1, *t1_proj.flatten(), t1_dist))
        print("Target 2: [{0:+06.2f}, {1:+06.2f}, {2:+06.2f}] ->"
              " [{3:+06.2f}, {4:+06.2f}, {5:+06.2f}] Dist: {6:05.2f}mm ".format(*target2, *t2_proj.flatten(), t2_dist))
        print(f"Dist1 - Dist2: {t1_dist - t2_dist:05.2f} mm")
        print(f"rMT Stokes corrected: {rmt_stokes:05.2f} %MSO")

    return rmt_stokes



import sys
import h5py
import os.path
import numpy as np


def render_coil_positions(
        coil_conf_set_1_positions,
        coil_conf_set_1_orientations,
        coil_conf_set_2_positions=None,
        coil_conf_set_2_orientations=None,
        coil_conf_set_3_positions=None,
        coil_conf_set_3_orientations=None,
        fn_mesh=None,
        surf_type="skin",
        viewport_dim=(1280, 720),
        camera_polar_coords=(-175, 66, 110),
        screenshot_fn=None,
        interactive=False):
    """
    Plots coil positions (and their orientation) in space,
    optionally also display a second set of coil positions,
    optionally also display a surface (e.g. gray matter or skin)

    Parameters:
    -----------
    coil_conf_set_1_positions : np.ndarray
        The "center" coordinates of the primary coil configurations
    coil_conf_set_1_orientations : np.ndarray
         Either the "m0", "m1" or "m2" orientation vector of the primary coil position.
    coil_conf_set_2_positions : np.ndarray | None, optional
        The "center" coordinates of the secondary coil configurations. Can be omitted.
    coil_conf_set_2_orientations : np.ndarray | None, optional
         Either the "m0", "m1" or "m2" orientation vector of the secondary coil position. Can be omitted.
    fn_mesh : str, optional
        Path to the mesh hdf5-file containing the geometry information of a head model. Can be omitted.
    surf_type : str, optional
        Name of the surface to display. Currently supported: "skin", "skull", "csf", "gm", "wm"
        Only valid if 'fn_mesh' has been provided.

    Returns:
    --------
    True if successful, False in case of a fatal error.
    """
    try:
        import mayavi.mlab
    except ModuleNotFoundError:
        print("[render_3D.py] Error: Cannot import module mayavi. Most rendering functions will not work.")
    if "mayavi" in sys.modules:
        if not interactive:
            mayavi.mlab.options.offscreen = True
        else:
            mayavi.mlab.options.offscreen = False

        if not isinstance(viewport_dim, tuple) or len(viewport_dim) != 2:
            viewport_dim = (1280, 720)
            print(
                "[render_coil_positions] Error: Provided viewport dimensions are not in the required format: (x,y),"
                "expected 2 integer representing x,y coordinates of the viewport. Will use default dimensions instead:"
                "(1280,720)")

        random_figure_id = int(np.random.default_rng().random(1)[0] * 1e9)
        fig = mayavi.mlab.figure(figure=random_figure_id, bgcolor=(1, 1, 1), engine=None, fgcolor=(0., 0., 0.),
                                 size=viewport_dim)

        if fn_mesh is not None and os.path.exists(fn_mesh):
            with h5py.File(fn_mesh, 'r') as mesh_h5:

                # tissue tags from SimNIBS v2 - v4
                tissue_tag = 1005  # default to skin

                if surf_type == "wm":
                    tissue_tag = 1001
                elif surf_type == "gm":
                    tissue_tag = 1002
                elif surf_type == "csf":
                    tissue_tag = 1003
                elif surf_type == "skull":
                    tissue_tag = 1007
                elif surf_type == "skin":
                    tissue_tag = 1005
                else:
                    print(f"[render_coil_positions] Error: Unsupported surface type '{surf_type}'. Will "
                          f"render skin surface per default.")

                mesh_tissue_type = np.array(mesh_h5["/mesh/elm/tri_tissue_type"])
                mesh_tris = np.array(mesh_h5["/mesh/elm/node_number_list"][:, :3])
                mesh_pts = np.array(mesh_h5["/mesh/nodes/node_coord"])
                surf_tris_idcs = np.where(mesh_tissue_type == tissue_tag)
                surf_tris = mesh_tris[surf_tris_idcs]

                scalp_surf = mayavi.mlab.triangular_mesh(
                    mesh_pts[:, 0],
                    mesh_pts[:, 1],
                    mesh_pts[:, 2],
                    surf_tris,
                    representation='surface',
                    opacity=1,
                    color=(0.8, 0.8, 0.8)
                )

        if coil_conf_set_2_positions is not None and coil_conf_set_2_orientations is not None:
            mayavi.mlab.points3d(
                coil_conf_set_2_positions[:, 0],
                coil_conf_set_2_positions[:, 1],
                coil_conf_set_2_positions[:, 2],
                scale_factor=0.3,
                color=(.9, .9, .9)
            )

            mayavi.mlab.quiver3d(
                coil_conf_set_2_positions[:, 0],
                coil_conf_set_2_positions[:, 1],
                coil_conf_set_2_positions[:, 2],
                coil_conf_set_2_orientations[:, 0],
                coil_conf_set_2_orientations[:, 1],
                coil_conf_set_2_orientations[:, 2],
                mode="arrow",
                scale_factor=1,
                color=(.9, .9, .9)
            )
        mayavi.mlab.points3d(
            coil_conf_set_1_positions[:, 0],
            coil_conf_set_1_positions[:, 1],
            coil_conf_set_1_positions[:, 2],
            scale_factor=0.5,
            color=(1., .0, .0)
        )

        mayavi.mlab.quiver3d(
            coil_conf_set_1_positions[:, 0],
            coil_conf_set_1_positions[:, 1],
            coil_conf_set_1_positions[:, 2],
            coil_conf_set_1_orientations[:, 0],
            coil_conf_set_1_orientations[:, 1],
            coil_conf_set_1_orientations[:, 2],
            mode="arrow",
            scale_factor=2,
            color=(1., .0, .0)
        )

        if coil_conf_set_3_positions is not None and coil_conf_set_3_orientations is not None:
            mayavi.mlab.points3d(
                coil_conf_set_3_positions[:, 0],
                coil_conf_set_3_positions[:, 1],
                coil_conf_set_3_positions[:, 2],
                scale_factor=2,
                color=(0.161, 0.812, 0.247)
            )

            mayavi.mlab.quiver3d(
                coil_conf_set_3_positions[:, 0],
                coil_conf_set_3_positions[:, 1],
                coil_conf_set_3_positions[:, 2],
                coil_conf_set_3_orientations[:, 0],
                coil_conf_set_3_orientations[:, 1],
                coil_conf_set_3_orientations[:, 2],
                mode="arrow",
                scale_factor=4,
                color=(0.161, 0.812, 0.247)
            )


        if not isinstance(camera_polar_coords, tuple) or len(camera_polar_coords) != 3:
            print(
                "[render_coil_positions] Error: Provided camera coordinates are not in the required format: (a,p,r), "
                "3 floats "
                "representing azimuthal angle, polar angle and radius. Will use default coordinates instead:"
                "(-175, 66, 110)")
            camera_polar_coords = (-175, 66, 110)

        mean_pts = np.mean(coil_conf_set_1_positions, axis=0)
        mayavi.mlab.view(
            azimuth=camera_polar_coords[0],
            elevation=camera_polar_coords[1],
            distance=camera_polar_coords[2],
            focalpoint=np.array(mean_pts),
            roll=None, reset_roll=None, figure=None
        )

        if screenshot_fn is not None:
            # Bug? Must execute the command twice to generate an output screenshot of the requested dimensions.
            # Otherwise its dimensions will be always (300,300).
            mayavi.mlab.savefig(filename=screenshot_fn, figure=fig)
            mayavi.mlab.savefig(filename=screenshot_fn, figure=fig)

        if interactive:
            mayavi.mlab.show()

        # In theory, passing a reference of the to-be closed figure should be
        # sufficient as well. In practice, it did not work reliably yielding
        # error after too many calls to that function because of too many open
        # figures. Introducing a (random) figure-ID and using this ID to refer
        # to the figure seemed to have solved the isse.
        mayavi.mlab.close(random_figure_id)
        mayavi.mlab.close(all=True)

        return True

    else:
        print("[render_coil_positions] Error: Required module 'mayavi' could be not successfully imported."
              "Is it installed? Cannot execute 'render_data_on_surface'.")

    return False


def render_data_on_surface(
        points,
        tris,
        data,
        viewport_dim=(1280, 720),
        camera_polar_coords=(-175, 66, 110),
        title=None,
        data_name="Data",
        colormap="jet",
        screenshot_fn=None,
        interactive=False
):
    """
    Plots data on surface:
    - If the number of data points equals the number of vertices in the mesh, the data will be displayed as point data.
    - If the number of data points equals the number of triangles in the mesh, the data will be displayed as cell data.

    Parameters
    ----------
    points : np.darray[float], [n_points x 3]
        Points (vertices) of surface mesh.
    tris : np.array[float], [n_tris x 3]
        Connectivity list of triangles.
    data : np.array[float], [n_tris]
        Data in triangular center
    viewport_dim : Tuple[int, int]
        Size of the viewport. Default: (1280, 720)
    camera_polar_coords : Tuple[float, float, float)
        The coordinate of the camera around the object in polar coordinates: (azimuthal angle, polar angle, radius)
    title : str
        Tile of the rendering window.
    data_name : str
        Name of the visualized data set.
    colormap : str
        Identifier of the desired color map. Available colormaps are:
        'Accent', 'Blues', 'BrBG', 'BuGn', 'BuPu', 'CMRmap', 'Dark2', 'GnBu', 'Greens', 'Greys'
        'OrRd', 'Oranges', 'PRGn', 'Paired', 'Pastel1', 'Pastel2', 'PiYG', 'PuBu', 'PuBuGn', 'PuOr',
        'PuRd', 'Purples', 'RdBu', 'RdGy', 'RdPu', 'RdYlBu', 'RdYlGn', 'Reds', 'Set1', 'Set2', 'Set3'
        'Spectral', 'Vega10', 'Vega20', 'Vega20b', 'Vega20c', 'Wistia', 'YlGn', 'YlGnBu', 'YlOrBr',
        'YlOrRd', 'afmhot', 'autumn', 'binary', 'black-white', 'blue-red', 'bone', 'brg', 'bwr',
        'cool', 'coolwarm', 'copper', 'cubehelix', 'file', 'flag', 'gist_earth', 'gist_gray',
        'gist_heat', 'gist_ncar', 'gist_rainbow', 'gist_stern', 'gist_yarg', 'gnuplot', 'gnuplot2', 'gray',
        'hot', 'hsv', 'inferno', 'jet', 'magma', 'nipy_spectral', 'ocean', 'pink', 'plasma', 'prism',
        'rainbow', 'seismic', 'spectral', 'spring', 'summer', 'terrain', 'viridis', 'winter'
    screenshot_fn : str | None
        If provided a screenshot will be saved to that path.
    interactive : bool
        If true, a blocking window will be spawned.

    Returns
    -------
    True if successful, False in case of a fatal error.
    """
    try:
        import mayavi.mlab
    except ModuleNotFoundError:
        print("[render_3D.py] Error: Cannot import module mayavi. Most rendering functions will not work.")
    if "mayavi" in sys.modules:
        if not interactive:
            mayavi.mlab.options.offscreen = True
        else:
            mayavi.mlab.options.offscreen = False

        if not isinstance(viewport_dim, tuple) or len(viewport_dim) != 2:
            viewport_dim = (1280, 720)
            print(
                "[render_data_on_surface] Error: Provided viewport dimensions are not in the required format: (x,y),"
                "expected 2 integer representing x,y coordinates of the viewport. Will use default dimensions instead:"
                "(1280,720)")

        random_figure_id = int(np.random.default_rng().random(1)[0] * 1e9)
        fig = mayavi.mlab.figure(figure=random_figure_id, bgcolor=(1, 1, 1), engine=None, fgcolor=(0., 0., 0.),
                                 size=viewport_dim)
        mesh = mayavi.mlab.triangular_mesh(points[:, 0], points[:, 1], points[:, 2], tris, representation='wireframe',
                                           opacity=0)

        if data.shape[0] == tris.shape[0]:
            mesh.mlab_source.dataset.cell_data.scalars = data
            mesh.mlab_source.dataset.cell_data.scalars.name = data_name
            mesh.mlab_source.update()
            mesh.parent.update()
            mesh = mayavi.mlab.pipeline.set_active_attribute(mesh, cell_scalars=data_name)
        elif data.shape[0] == points.shape[0]:
            mesh.mlab_source.dataset.point_data.scalars = data
            mesh.mlab_source.dataset.point_data.scalars.name = data_name
            mesh.mlab_source.update()
            mesh.parent.update()
            mesh = mayavi.mlab.pipeline.set_active_attribute(mesh, point_scalars=data_name)
        else:
            print(
                "[render_data_on_surface] Error: Provided data does neither equal the number of ertices nor triangles "
                "of the "
                "surface mesh. Cannot map data onto mesh.")
            return False

        surf = mayavi.mlab.pipeline.surface(mesh, colormap=colormap, vmin=0, vmax=1)

        if not isinstance(camera_polar_coords, tuple) or len(camera_polar_coords) != 3:
            print(
                "[render_data_on_surface] Error: Provided camera coordinates are not in the required format: (a,p,r), "
                "3 floats "
                "representing azimuthal angle, polar angle and radius. Will use default coordinates instead:"
                "(-175, 66, 110)")
            camera_polar_coords = (-175, 66, 110)

        mean_pts = np.mean(points, axis=0)
        mayavi.mlab.view(
            azimuth=camera_polar_coords[0],
            elevation=camera_polar_coords[1],
            distance=camera_polar_coords[2],
            focalpoint=np.array(mean_pts),
            roll=None, reset_roll=None, figure=None
        )

        mayavi.mlab.colorbar(
            surf,
            title=title
        )

        if screenshot_fn is not None:
            # Bug? Must execute the command twice to generate an output screenshot of the requested dimensions.
            # Otherwise its dimensions will be always (300,300).
            mayavi.mlab.savefig(filename=screenshot_fn, figure=fig)
            mayavi.mlab.savefig(filename=screenshot_fn, figure=fig)

        if interactive:
            mayavi.mlab.show()

        # In theory, passing a reference of the to-be closed figure should be
        # sufficient as well. In practice, it did not work reliably yielding
        # error after too many calls to that function because of too many open
        # figures. Introducing a (random) figure-ID and using this ID to refer
        # to the figure seemed to have solved the isse.
        mayavi.mlab.close(random_figure_id)

        return True
    else:
        print("[render_data_on_surface] Error: Required module 'mayavi' could not be successfully imported."
              "Is it installed? Cannot execute 'render_data_on_surface'.")

        return False



import numpy as np


def biphasic_pulse(t, R=0.0338, L=15.5*1e-6, C=193.6*1e-6, alpha=1089.8, f=2900):
    """
    Returns normalized single biphasic pulse waveform of electric field (first derivative of coil current)

    Parameters
    ----------
    t: ndarray of float [n_t]
        Time array in seconds
    R: float, optional, default: 0.0338 Ohm
        Resistance of coil in (Ohm)
    L: float, optional, default: 15.5*1e-6 H
        Inductance of coil in (H)
    C: float, optional, default: 193.6*1e-6
        Capacitance of coil in (F)
    alpha: float, optional, default: 1089.8 1/s
        Damping coefficient in (1/s)
    f: float, optional, default: 2900 Hz
        Frequency in (Hz)

    Returns
    -------
    e: ndarray of float [n_t]
        Normalized electric field time course (can be scaled with electric field)
    """

    omega = 2 * np.pi * f
    i = 1/(omega*L) * np.exp(-alpha*t) * np.sin(omega*t)
    e = np.gradient(i)
    e = e/np.max(e)

    return i, e



"""
Classes to cope with cortical region of interests (ROIs).
"""
import os
import math
import scipy
import skimage
import tqdm
import trimesh
import warnings
import numpy as np
import nibabel as nib
import scipy.interpolate

try:
    import simnibs
except ModuleNotFoundError:
    pass

import pynibs


class CorticalLayer:
    __create_key = object()

    class Settings:
        ROI_SIZE_OFFSET = 5
        GRID_POINTS_PER_MM = 1.5
        TAG_WHITE_MATTER_VOL = 1
        TAG_GRAY_MATTER_VOL = 2
        TAG_WHITE_MATTER_SURF = 1001
        TAG_GRAY_MATTER_SURF = 1002
        NUM_TRIANGLE_SMOOTHING_STEPS = 30

    def __init__(self, create_key, layer_id, volumetric_mesh=None, roi=None, depth=None, path=None, surface=None,
                 id=None):
        """
        Constructor of the cortical layer class. Three optional ways of construction a CorticalLayer-instance
        a) by providing a path to an already existing layer (simnibs.Msh)
        b) by proving a simnibs.Msh of an already existing layer.
        c) by providing the bounding box of the ROI in which the layer should be created.

        Parameters
        ----------
        layer_id : str
            Identifier of the layer.
        volumetric_mesh : simnibs.Msh, optional
            The tetrahedral volume mesh, in which the layer should be generated.
        roi : pynibs.RegionOfInterestSurface
            RegionOfInterestSurface.
        depth: float, optional
            Normalized distance of the layer from gray matter surface.
            Provide values in the open interval (0,1).
        path : str, optional
            File path to a region of interest surfe (e.g. midlayer).
        surface : simnibs.Msh, optional
            The surface representation of an already existing layer (e.g. midlayer).
        id : str, optional, deprecated
            This was replaced by layer_id.
        """
        if id is not None:
            warnings.warn(DeprecationWarning("The parameter 'id' is deprecated and will be removed in future versions. "
                                             "use 'layer_id' instead."))
            assert layer_id is None, "Please use 'layer_id' instead of 'id'."
            layer_id = id

        # trying to mimic a private constructor here, from: https://stackoverflow.com/a/46459300
        assert create_key == CorticalLayer.__create_key, \
            "Direct construction not allowed. Please use factory methods " \
            "'init_from_file', 'init_from_surface', 'create_in_bbox'"
        self.id = layer_id

        CorticalLayer.Settings.ROI_SIZE_OFFSET = 5

        if path is not None:
            self.surface = simnibs.read_msh(path)
            self.roi = CorticalLayer.roi_bbox_from_points(self.surface.nodes.node_coord,
                                                          CorticalLayer.Settings.ROI_SIZE_OFFSET)
        elif surface is not None:
            self.surface = surface
            self.roi = CorticalLayer.roi_bbox_from_points(self.surface.nodes.node_coord,
                                                          CorticalLayer.Settings.ROI_SIZE_OFFSET)
        elif roi is not None and depth is not None:
            bbox = [
                np.min(roi.node_coord_mid[:, 0]),
                np.max(roi.node_coord_mid[:, 0]),
                np.min(roi.node_coord_mid[:, 1]),
                np.max(roi.node_coord_mid[:, 1]),
                np.min(roi.node_coord_mid[:, 2]),
                np.max(roi.node_coord_mid[:, 2])
            ]

            self.volumetric_mesh = CorticalLayer.crop_mesh_with_box(
                volumetric_mesh,
                bbox,
                True
            )

            self.surface = None
            self.roi = roi
            self.generate_layer(depth, roi)
        else:
            raise ValueError('At least one set of optional parameters must be assigned')

    @classmethod
    def init_from_file(cls, layer_id, fn):
        """
        Factory method for constructing a CorticalLayer-object from a file.

        Parameters
        ----------
        layer_id : str
            Identifier of the layer.
        fn : str
            File path to a region of interest surfe (e.g. midlayer).
        """
        return CorticalLayer(cls.__create_key, layer_id=layer_id, path=fn)

    @classmethod
    def init_from_surface(cls, layer_id, surf):
        """
        Factory method for constructing a CorticalLayer-object from a Simnibs-surface object.

        Parameters
        ----------
        layer_id : str
            Identifier of the layer.
        surf : simnibs.Msh
            The surface representation of an already existing layer (e.g. midlayer).
        """
        return CorticalLayer(cls.__create_key, layer_id=layer_id, surface=surf)

    @classmethod
    def create_in_roi(cls, layer_id, roi, depth, volmesh):
        """
        Factory method for constructing a CorticalLayer-object
        within a region-of-interest and a specified cortical depth.

        Parameters
        ----------
        layer_id : str
            Identifier of the layer.
        roi : pynibs.RegionOfInterestSurface
            RegionOfInterestSurface.
        depth: float
            Normalized distance of the layer from gray matter surface.
            Provide values in the open interval (0,1).
        volmesh : simnibs.Msh
            The tetrahedral volume mesh, in which the layer should be generated.
        """
        return CorticalLayer(cls.__create_key, layer_id=layer_id, volumetric_mesh=volmesh, roi=roi, depth=depth)

    @classmethod
    def create_in_bbox(cls, layer_id, bbox, depth, volmesh):
        """
        Factory method for constructing a CorticalLayer-object
        within a region-of-interest and a specified cortical depth.

        Parameters
        ----------
        layer_id : str
            Identifier of the layer.
        bbox : typing.List[float]
            List of bounding values around the ROI box: [x_min, x_max, y_min, y_max, z_min, z_max].
        depth: float
            Normalized distance of the layer from gray matter surface.
            Provide values in the open interval (0,1).
        volmesh : simnibs.Msh
            The tetrahedral volume mesh, in which the layer should be generated.
        """
        return CorticalLayer(cls.__create_key, layer_id=layer_id, volumetric_mesh=volmesh, roi=bbox, depth=depth)

    @staticmethod
    def roi_bbox_from_points(points, offset=0):
        """
        Find the minimal bounding box around the provided points.

        Parameters
        ----------
        points : simnibs.Msh
            The tetrahedral volume mesh, in which the layer should be generated.
        offset: float
            Normalized distance of the layer from gray matter surface.
            Provide values in the open interval (0,1).

        Returns
        -------
        bounding_box : typing.List[float]
            List of bounding values around the provided points: [x_min, x_max, y_min, y_max, z_min, z_max].
        """
        return [points[:, 0].min() - offset, points[:, 0].max() + offset,
                points[:, 1].min() - offset, points[:, 1].max() + offset,
                points[:, 2].min() - offset, points[:, 2].max() + offset]

    @staticmethod
    def crop_mesh_with_box(mesh, roi, keep_elements=False):
        """
        Returns the cropped mesh with all points that are inside the region of interest

        Parameters
        ----------
        keep_elements : bool, default = False
            If True, keeps elements with at least one point in roi, else removes them.
        mesh: simnibs.Msh
             The mesh that is supposed to be cropped.
        roi: typing.List[float]
             The bounding box of the region of interest which the mesh should be cropped to.
             [x-min, x-max, y-min, y-max, z-min, z-max].

        Returns
        -------
        mesh_cropped : simnibs.Msh
            The cropped mesh.
        """
        node_keep_indexes = np.where(
            np.all(
                np.logical_and(
                    mesh.nodes.node_coord <= [roi[1], roi[3], roi[5]],
                    [roi[0], roi[2], roi[4]] <= mesh.nodes.node_coord
                ),
                axis=1
            )
        )[0] + 1

        if keep_elements:  # crop using node-based indices
            return mesh.crop_mesh(nodes=node_keep_indexes)
        else:  # crop using element-based indices
            elements_to_keep = np.where(
                np.all(
                    np.isin(
                        mesh.elm.node_number_list,
                        node_keep_indexes
                    ).reshape(-1, 4),
                    axis=1
                )
            )[0] + 1

            return mesh.crop_mesh(elements=elements_to_keep)

    @staticmethod
    def crop_mesh_with_surface(mesh, roi, keep_elements=False, radius=3):
        """
        Returns the cropped mesh with all points that are close to the surface of interest

        Parameters
        ----------
        keep_elements : bool, default = False
            If True, keeps elements with at least one point in roi, else removes them.
        mesh: simnibs.Msh
             The mesh that is supposed to be cropped.
        roi: RegionOfInterestSurface instance
             RegionOfInterestSurface.
        radius : float, default = 3
            Search radius of mesh elements around ROI nodes.

        Returns
        -------
        mesh_cropped : simnibs.Msh
            The cropped mesh.
        """
        node_keep = np.zeros(mesh.nodes.node_coord.shape[0]).astype(bool)

        for vertex in roi.node_coord_mid:
            node_keep = np.logical_or(node_keep,
                                      np.linalg.norm(mesh.nodes.node_coord - vertex, axis=1) < radius)
        node_keep_indexes = np.where(node_keep)[0] + 1

        if keep_elements:  # crop using node-based indices
            return mesh.crop_mesh(nodes=node_keep_indexes)
        else:  # crop using element-based indices
            elements_to_keep = np.where(
                np.all(
                    np.isin(
                        mesh.elm.node_number_list,
                        node_keep_indexes
                    ).reshape(-1, 4),
                    axis=1
                )
            )[0] + 1

            return mesh.crop_mesh(elements=elements_to_keep)

    def generate_layer(self, depth, roi):
        """
        Create the geometry of the layer at the specified depth using marching cubes.

        Parameters
        ----------
        depth : float
            The depth below the GM surface at which the layer should be generated; in [0,1].
        roi : RegionOfInterestSurface instance
            RegionOfInterestSurface.
        """
        bbox = [
            np.min(roi.node_coord_mid[:, 0]),
            np.max(roi.node_coord_mid[:, 0]),
            np.min(roi.node_coord_mid[:, 1]),
            np.max(roi.node_coord_mid[:, 1]),
            np.min(roi.node_coord_mid[:, 2]),
            np.max(roi.node_coord_mid[:, 2])
        ]

        roi_extended = [
            bbox[0] - CorticalLayer.Settings.ROI_SIZE_OFFSET,
            bbox[1] + CorticalLayer.Settings.ROI_SIZE_OFFSET,
            bbox[2] - CorticalLayer.Settings.ROI_SIZE_OFFSET,
            bbox[3] + CorticalLayer.Settings.ROI_SIZE_OFFSET,
            bbox[4] - CorticalLayer.Settings.ROI_SIZE_OFFSET,
            bbox[5] + CorticalLayer.Settings.ROI_SIZE_OFFSET
        ]

        # 1) Create grid points used for interpolation.
        grid_x = np.linspace(roi_extended[0], roi_extended[1],
                             int(math.fabs(
                                 roi_extended[0] - roi_extended[1]) * CorticalLayer.Settings.GRID_POINTS_PER_MM))
        grid_y = np.linspace(roi_extended[2], roi_extended[3],
                             int(math.fabs(
                                 roi_extended[2] - roi_extended[3]) * CorticalLayer.Settings.GRID_POINTS_PER_MM))
        grid_z = np.linspace(roi_extended[4], roi_extended[5],
                             int(math.fabs(
                                 roi_extended[4] - roi_extended[5]) * CorticalLayer.Settings.GRID_POINTS_PER_MM))
        grid_points = np.stack(np.meshgrid(grid_x, grid_y, grid_z, indexing='ij'), axis=-1).reshape(-1, 3)

        # 2) Find points in interpolation grid that are inside/outside the gray matter.
        tet_idcs = self.volumetric_mesh.find_tetrahedron_with_points(grid_points, compute_baricentric=False)

        point_indices_in_volume = np.where(tet_idcs != -1)[0]

        point_tissue_tag = np.ones(tet_idcs.shape) * -1

        point_tissue_tag[point_indices_in_volume] = self.volumetric_mesh.elm.tag1[
            tet_idcs[point_indices_in_volume] - 1  # make indices 0-based
            ]
        grid_point_idcs_outside_gm_wm = np.where(
            (point_tissue_tag != CorticalLayer.Settings.TAG_WHITE_MATTER_VOL)
            &
            (point_tissue_tag != CorticalLayer.Settings.TAG_GRAY_MATTER_VOL)
        )[0]
        grid_point_idcs_inside_gm = np.where(point_tissue_tag == CorticalLayer.Settings.TAG_GRAY_MATTER_VOL)[0]

        # Define vertices of the ROI bounding box and associated data.
        bounding_roi_pts = np.stack(
            np.meshgrid(roi_extended[0:2], roi_extended[2:4], roi_extended[4:6], indexing='ij'),
            axis=-1
        ).reshape(-1, 3)

        bbox_tet_idcs = self.volumetric_mesh.find_tetrahedron_with_points(bounding_roi_pts, compute_baricentric=False)
        bounding_roi_pts_tissue_tags = self.volumetric_mesh.elm.tag1[
            bbox_tet_idcs - 1  # make indices 0-based
            ]
        # init interpolation points with 1 for WM and 0 outside WM
        bounding_roi_pts_init_vals = np.ones(bounding_roi_pts_tissue_tags.shape)
        # max non-WM bounding box points
        bounding_roi_pts_init_vals[bounding_roi_pts_tissue_tags != CorticalLayer.Settings.TAG_WHITE_MATTER_VOL] = 0
        # if bounding box point is located outside the volume mesh, treat it as non-WM as well (even if it is WM tissue)
        bounding_roi_pts_init_vals[np.where(bbox_tet_idcs == -1)] = 0

        # 3) Prepare interpolator and interpolate on grid points that are inside GM.
        wm_surface_nodes = self.volumetric_mesh.crop_mesh(CorticalLayer.Settings.TAG_WHITE_MATTER_SURF).nodes.node_coord
        gm_surface_nodes = self.volumetric_mesh.crop_mesh(CorticalLayer.Settings.TAG_GRAY_MATTER_SURF).nodes.node_coord

        # Create a gradient (from 0 to 1) between GM and WM.
        data = [1] * len(wm_surface_nodes)  # init interpolation points with 1 for grid points inside WM
        data += [0] * len(gm_surface_nodes)  # init interpolation points with 0 for grid points inside GM
        # use interpolation points at bbox as "outside points" to ensure
        # there is always a defined interpolation point also outside the GM volume
        data += bounding_roi_pts_init_vals.tolist()

        data_points = np.concatenate((wm_surface_nodes, gm_surface_nodes, bounding_roi_pts), axis=0)

        interpolation = scipy.interpolate.LinearNDInterpolator(data_points, data, fill_value=-1)
        gray_matter_interpolation = interpolation(grid_points[grid_point_idcs_inside_gm])

        # 4) Marching-cubes-based surface creation.
        volume_data = np.empty((len(grid_x), len(grid_y), len(grid_z)))
        volume_data.fill(1)
        outside_x = (grid_point_idcs_outside_gm_wm / (len(grid_z) * len(grid_y))).astype(int)
        outside_y = ((grid_point_idcs_outside_gm_wm / len(grid_z)) % len(grid_y)).astype(int)
        outside_z = (grid_point_idcs_outside_gm_wm % len(grid_z)).astype(int)
        volume_data[(outside_x, outside_y, outside_z)] = 0
        inside_gray_matter_x = (grid_point_idcs_inside_gm / (len(grid_z) * len(grid_y))).astype(int)
        inside_gray_matter_y = ((grid_point_idcs_inside_gm / len(grid_z)) % len(grid_y)).astype(int)
        inside_gray_matter_z = (grid_point_idcs_inside_gm % len(grid_z)).astype(int)
        volume_data[(inside_gray_matter_x, inside_gray_matter_y, inside_gray_matter_z)] = gray_matter_interpolation
        vertices, faces, _, _ = skimage.measure.marching_cubes(
            volume_data,
            level=depth,
            spacing=tuple(
                np.array(
                    [grid_x[1] - grid_x[0], grid_y[1] - grid_y[0], grid_z[1] - grid_z[0]],
                    dtype='float32'
                )
            ),
            step_size=1,
            allow_degenerate=False
        )

        # 5) prepare surface for output
        self.surface = simnibs.Msh(
            simnibs.Nodes(vertices),
            simnibs.Elements(faces + 1)
        )
        self.remove_unconnected_surfaces()
        self.surface.nodes.node_coord = self.surface.nodes.node_coord + [(roi_extended[0]), (roi_extended[2]),
                                                                         (roi_extended[4])]

        # Crop again for a more precise crop-boundary (earlier cropping was done with larger elements)
        self.surface = CorticalLayer.crop_mesh_with_box(self.surface, bbox, keep_elements=True)
        self.surface = CorticalLayer.crop_mesh_with_surface(self.surface, roi, keep_elements=True)
        self.remove_unconnected_surfaces()

        # from simnibs.Msh.fix_surface_orientation (not implemented in SimNIBS < v.4)
        idx_tr = self.surface.elm.elm_type == 2
        normals = self.surface.triangle_normals()[:]
        baricenters = self.surface.elements_baricenters()[idx_tr]
        CoG = np.mean(baricenters, axis=0)

        nr_inward = sum(np.einsum("ij,ij->i", normals, baricenters - CoG) < 0)

        if nr_inward / sum(idx_tr) > 0.5:
            buffer = self.surface.elm.node_number_list[idx_tr, 1].copy()
            self.surface.elm.node_number_list[idx_tr, 1] = self.surface.elm.node_number_list[idx_tr, 2]
            self.surface.elm.node_number_list[idx_tr, 2] = buffer

    def get_smoothed_normals(self):
        """
        Computed the smoothed normals of the surface representation of this layer.

        Note: For the later stages, we don't want a smoothed surface, but smooth
        normals in order to maintain the location of the cells, but orient
        them more smoothly. Therefore, we use smoothed normals, e.g. for the
        computation of the theta angle, but do not smooth the entire layer
        surface.

        Returns
        -------
        normals : np.ndarray
            The tetrahedral volume mesh, in which the layer should be generated.
        """
        return self.surface.triangle_normals(smooth=CorticalLayer.Settings.NUM_TRIANGLE_SMOOTHING_STEPS).value

    def save(self, fn):
        """
        Save the current surface representation of this CorticalLayer instance at the specified location.

        Parameters
        ----------
        fn : str
            Target file name of the surface-file of this layer.
        """
        self.surface.write(fn)

    def remove_unconnected_surfaces(self):
        """
        Remove elements small unconnected element-clusters from this layer.
        """
        surfaces = self.surface.elm.connected_components()
        surfaces.sort(key=len)
        self.surface = self.surface.crop_mesh(elements=surfaces[-1])

    def get_evenly_spaced_element_subset(self, elements_per_square_mm):
        """
        Subsample the surface representation of the ayer.

        Parameters
        ----------
        elements_per_square_mm : float
            Number of triangles per mm^2 in the layer.

        Returns
        -------
        selected elements : Typing.List[int]
            List of indices of selected elements as a result of the subsampling.
        """
        centers = self.surface.elements_baricenters().value
        min_distance_square = (1 / elements_per_square_mm) * math.sqrt(2) / 2  # = radius of circumference
        selected_elements = np.array([0])

        # For each element: add to the list of 'selected_elements' if the distance to the already
        # selected elements is larger than the minimum element density.
        for element_index in range(self.surface.elm.nr):
            selected_elements_centers = centers[selected_elements]
            element_center = centers[element_index]
            distances_square = (selected_elements_centers[:, 0] - element_center[0]) ** 2 + \
                               (selected_elements_centers[:, 1] - element_center[1]) ** 2 + \
                               (selected_elements_centers[:, 2] - element_center[2]) ** 2
            if distances_square.min() > min_distance_square:
                selected_elements = np.append(selected_elements, element_index)

        return np.array(selected_elements)


class RegionOfInterestSurface:
    """
    Region of interest (surface).

    Attributes
    ----------
    node_coord_up : np.ndarray
        (N_points, 3) Coordinates (x,y,z) of upper surface nodes.
    node_coord_mid : np.ndarray
        (N_points, 3) Coordinates (x,y,z) of middle surface nodes.
    node_coord_low : np.ndarray
        (N_points, 3) Coordinates (x,y,z) of lower surface nodes.
    node_number_list : np.ndarray
        (N_points, 3) Connectivity matrix of triangles.
    delta : float
        Distance parameter between WM and GM (0 -> WM, 1 -> GM).
    tet_idx_tri_center_up : np.ndarray [N_points]
        Tetrahedra indices of TetrahedraLinear object instance where the center points of the triangles of the
        upper surface are.
    tet_idx_tri_center_mid : np.ndarray [N_points]
        Tetrahedra indices of TetrahedraLinear object instance where the center points of the triangles of the
        middle surface are.
    tet_idx_tri_center_low : np.ndarray [N_points]
        Tetrahedra indices of TetrahedraLinear object instance where the center points of the triangles of the
        lower surface are.
    tet_idx_node_coord_mid : np.ndarray
        (N_tri,) Tetrahedra indices of TetrahedraLinear object instance where the nodes of the middle surface are.
    tri_center_coord_up : np.ndarray
        (N_tri, 3) Coordinates of roi triangle center of upper surface
    tri_center_coord_mid : np.ndarray
        (N_tri, 3) Coordinates of roi triangle center of middle surface
    tri_center_coord_low : np.ndarray
        (N_tri, 3) Coordinates of roi triangle center of lower surface
    fn_mask : string
        Filename for surface mask in subject space. .mgh file or freesurfer surface file.
    fn_mask_avg : string
        Filename for .mgh mask in fsaverage space. Absolute path or relative to mesh folder.
    fn_mask_nii : string
        Filename for .nii or .nii.gz mask. Absolute path or relative to mesh folder.
    X_ROI : list of float
        Region of interest [Xmin, Xmax], whole X range if empty [0,0] or None
        (left - right)
    Y_ROI : list of float
        Region of interest [Ymin, Ymax], whole Y range if empty [0,0] or None
        (anterior - posterior)
    Z_ROI : list of float
        Region of interest [Zmin, Zmax], whole Z range if empty [0,0] or None
        (inferior - superior)
    template : str
        'MNI', 'fsaverage', 'subject'
    center : list of float
        Center coordinates for spherical ROI in self.template space
    radius : float
        Radius in [mm] for spherical ROI
    gm_surf_fname : str or list of str
        Filename(s) of GM surface generated by freesurfer (lh and/or rh)
        (e.g. in mri2msh: .../fs_ID/surf/lh.pial)
    wm_surf_fname : str or list of str
        Filename(s) of WM surface generated by freesurfer (lh and/or rh)
        (e.g. in mri2msh: .../fs_ID/surf/lh.white)
    layer : int
        Define the number of layers:

        * 1: one layer
        * 3: additionally upper and lower layers are generated around the central midlayer
    """
    def __init__(self):
        """
        Initialize RegionOfInterestSurface class instance
        """

        self.node_coord_up = np.empty(0)
        self.node_coord_mid = np.empty(0)
        self.node_coord_low = np.empty(0)

        self.node_number_list = np.empty(0)
        self.delta = []

        self.tet_idx_tri_center_up = np.empty(0)
        self.tet_idx_tri_center_mid = np.empty(0)
        self.tet_idx_tri_center_low = np.empty(0)

        self.tet_idx_node_coord_mid = np.empty(0)

        self.tri_center_coord_up = []
        self.tri_center_coord_mid = []
        self.tri_center_coord_low = []

        self.template = None
        self.fn_mask = []
        self.fn_mask_avg = None
        self.fn_mask_nii = None

        self.X_ROI = []
        self.Y_ROI = []
        self.Z_ROI = []

        self.center = None
        self.radius = None

        self.gm_surf_fname = []
        self.wm_surf_fname = []
        self.midlayer_surf_fname = []
        self.layer = []
        self.mesh_folder = []
        self.refine = []

        self.n_tris = -1
        self.n_nodes = -1
        self.n_tets = -1

        self.layers = []

    def project_on_midlayer(self, target, verbose=False):
        """
        Project a coordinate on the nearest midlayer node

        Parameters
        ----------
        target : np.ndarray
            Coordinate to project as  (3,) array
        verbose : bool
            Print some verbosity information. Default: False

        Returns
        -------
        target_proj : np.ndarray
            Node coordinate of nearest midlayer node.
        """
        # find midlayer node that is nearest to the target
        target_node_coord_mid = np.where(np.linalg.norm(self.node_coord_mid - target, axis=1) == np.min(
            np.linalg.norm(self.node_coord_mid - target, axis=1)))[0][0]

        # get coordinates of that node
        target_proj = self.node_coord_mid[target_node_coord_mid]

        if verbose:
            print(f"Projected {target} to {target_proj} (Dist: {np.linalg.norm(target - target_proj):2.2f}mm)")

        return target_proj

    def make_GM_WM_surface(self, gm_surf_fname=None, wm_surf_fname=None, midlayer_surf_fname=None, mesh_folder=None,
                           delta=0.5,
                           x_roi=None, y_roi=None, z_roi=None,
                           layer=1,
                           fn_mask=None, refine=False):
        """
        Generating a surface between WM and GM in a distance of delta 0...1 for ROI,
        given by Freesurfer mask or coordinates.

        Parameters
        ----------
        gm_surf_fname : str or list of str
            Filename(s) of GM FreeSurfer surface(s)  (lh and/or rh).
            Either relative to mesh_folder (fs_ID/surf/lh.pial) or absolute (/full/path/to/lh.pial)
        wm_surf_fname : str or list of str
            Filename(s) of WM FreeSurfer surface(s) (lh and/or rh)
            Either relative to mesh_folder (fs_ID/surf/lh.white) or absolute (/full/path/to/lh.white)
        midlayer_surf_fname : str or list of str
            Filename(s) of midlayer surface (lh and/or rh)
            Either relative to mesh_folder (fs_ID/surf/lh.central) or absolute (/full/path/to/lh.central)
        mesh_folder : str
            Root folder of mesh, Needed if paths above are given relative, or refine=True
        [defunct] m2m_mat_fname : str
            Filename of mri2msh transformation matrix
            (e.g. in mri2msh: .../m2m_ProbandID/MNI2conform_6DOF.mat)
        delta : float
            Distance parameter where surface is generated 0...1 (default: 0.5)

            * 0 -> WM surface
            * 1 -> GM surface
        x_roi : list of float
            Region of interest [Xmin, Xmax], whole X range if empty [0,0] or None
            (left - right)
        y_roi : list of float
            Region of interest [Ymin, Ymax], whole Y range if empty [0,0] or None
            (anterior - posterior)
        z_roi : list of float
            Region of interest [Zmin, Zmax], whole Z range if empty [0,0] or None
            (inferior - superior)
        layer : int
            Define the number of layers:

            * 1: one layer
            * 3: additionally upper and lower layers are generated around the central midlayer
        fn_mask : str
            Filename for FreeSurfer .mgh mask.
        refine : bool, optional, default: False
            Refine ROI by splitting elements

        Returns
        -------
        node_coord_up : np.ndarray of float [N_roi_points x 3]
            Node coordinates (x, y, z) of upper epsilon layer of ROI surface
        node_coord_mid : np.ndarray of float [N_roi_points x 3]
            Node coordinates (x, y, z) of ROI surface
        node_coord_low : np.ndarray of float [N_roi_points x 3]
            Node coordinates (x, y, z) of lower epsilon layer of ROI surface
        node_number_list : np.ndarray of int [N_roi_tri x 3]
            Connectivity matrix of intermediate surface layer triangles
        delta : float
            Distance parameter where surface is generated 0...1 (default: 0.5)

            * 0 -> WM surface
            * 1 -> GM surface
        tri_center_coord_up : np.ndarray of float [N_roi_tri x 3]
            Coordinates (x, y, z) of triangle center of upper epsilon layer of ROI surface
        tri_center_coord_mid : np.ndarray of float [N_roi_tri x 3]
            Coordinates (x, y, z) of triangle center of ROI surface
        tri_center_coord_low : np.ndarray of float [N_roi_tri x 3]
            Coordinates (x, y, z) of triangle center of lower epsilon layer of ROI surface
        fn_mask : str
            Filename for freesurfer mask. If given, this is used instead of *_ROIs
        X_ROI : list of float
            Region of interest [Xmin, Xmax], whole X range if empty [0,0] or None
            (left - right)
        Y_ROI : list of float
            Region of interest [Ymin, Ymax], whole Y range if empty [0,0] or None
            (anterior - posterior)
        Z_ROI : list of float
            Region of interest [Zmin, Zmax], whole Z range if empty [0,0] or None
            (inferior - superior)

        Example
        -------
        .. code-block:: python

            make_GM_WM_surface(self, gm_surf_fname, wm_surf_fname, delta, X_ROI, Y_ROI, Z_ROI)
            make_GM_WM_surface(self, gm_surf_fname, wm_surf_fname, delta, mask_fn, layer=3)
        """
        self.gm_surf_fname = gm_surf_fname
        self.wm_surf_fname = wm_surf_fname
        self.midlayer_surf_fname = midlayer_surf_fname
        self.layer = layer
        self.mesh_folder = mesh_folder
        self.fn_mask = fn_mask
        self.delta = delta
        self.X_ROI = x_roi
        self.Y_ROI = y_roi
        self.Z_ROI = z_roi
        self.refine = refine

        if type(gm_surf_fname) is not list:
            gm_surf_fname = [gm_surf_fname]

        if type(wm_surf_fname) is not list:
            wm_surf_fname = [wm_surf_fname]

        if type(midlayer_surf_fname) is not list:
            midlayer_surf_fname = [midlayer_surf_fname]

        if len(gm_surf_fname) != len(wm_surf_fname):
            raise ValueError('provide equal number of GM and WM surfaces!')

        # load surface data
        points_gm = [None for _ in range(len(gm_surf_fname))]
        points_wm = [None for _ in range(len(wm_surf_fname))]
        points_mid = [None for _ in range(len(midlayer_surf_fname))]
        con_gm = [None for _ in range(len(gm_surf_fname))]
        con_mid = [None for _ in range(len(midlayer_surf_fname))]

        max_idx_gm = 0
        max_idx_mid = 0

        def read_geom(fn, max_idx):
            """
            Read freesurfer geometries, either in .central format or as .gii
            """
            if not fn.startswith(os.sep):
                fn = os.path.join(mesh_folder, fn)

            # charm uses .gii files
            if fn.endswith('.gii'):
                # FIX: add_data returns DataArrays in the order of appearance in the file
                #      This does not necessarily have to be first points then triangles.
                #      So "points, con = nib.load(fn).agg_data()" may lead to swapped points and con.
                con = nib.load(fn).agg_data('NIFTI_INTENT_TRIANGLE')
                points = nib.load(fn).agg_data('NIFTI_INTENT_POINTSET')

            # headreco uses .central files
            else:
                points, con = nib.freesurfer.read_geometry(fn)

            con += max_idx
            max_idx += points.shape[0]  # np.max(con_gm[i]) + 2
            return points, con, max_idx

        for i in range(len(gm_surf_fname)):
            if gm_surf_fname[i] is not None:
                points_gm[i], con_gm[i], max_idx_gm = read_geom(gm_surf_fname[i], max_idx_gm)

            if wm_surf_fname[i] is not None:
                points_wm[i], _, max_idx_wm = read_geom(wm_surf_fname[i], max_idx_gm)

            if midlayer_surf_fname[i] is not None:
                points_mid[i], con_mid[i], max_idx_mid = read_geom(midlayer_surf_fname[i], max_idx_mid)

        points_gm = np.vstack(points_gm)
        points_wm = np.vstack(points_wm)
        points_mid = np.vstack(points_mid)
        con_gm = np.vstack(con_gm)
        con_mid = np.vstack(con_mid)

        # Determine 3 layer midlayer if GM and WM surfaces are present otherwise use provided midlayer data
        if gm_surf_fname[0] is not None and wm_surf_fname[0] is not None:
            # determine vector pointing from wm surface to gm surface
            wm_gm_vector = points_gm - points_wm

            eps_0 = 0.025
            eps, surface_points_upper, surface_points_lower = (False,) * 3
            surface_points_upper = False
            if layer == 3:
                # set epsilon range for upper and lower surface
                if delta < eps_0:
                    eps = delta / 2
                elif delta > (1 - eps_0):
                    eps = (1 - delta) / 2
                else:
                    eps = eps_0

            # determine wm-gm surfaces
            surface_points_middle = points_wm + wm_gm_vector * delta  # type: np.ndarray
            if layer == 3:
                surface_points_upper = surface_points_middle + wm_gm_vector * eps
                surface_points_lower = surface_points_middle - wm_gm_vector * eps

            con = con_gm

        elif midlayer_surf_fname[0] is not None:
            self.layer = 1

            surface_points_upper = None
            surface_points_lower = None
            surface_points_middle = points_mid

            con = con_mid

        else:
            raise IOError("Please provide GM and WM surfaces or midlayer "
                          "surface directly for midlayer calculation...")

        # crop region if desired
        x_roi_wb, y_roi_wb, z_roi_wb = (False,) * 3
        if fn_mask is None:
            # crop to region of interest
            if x_roi is None or x_roi == [0, 0]:
                x_roi = [-np.inf, np.inf]
                x_roi_wb = True
            if y_roi is None or y_roi == [0, 0]:
                y_roi = [-np.inf, np.inf]
                y_roi_wb = True
            if z_roi is None or z_roi == [0, 0]:
                z_roi = [-np.inf, np.inf]
                z_roi_wb = True

            roi_mask_bool = (surface_points_middle[:, 0] > min(x_roi)) & (surface_points_middle[:, 0] < max(x_roi)) & \
                            (surface_points_middle[:, 1] > min(y_roi)) & (surface_points_middle[:, 1] < max(y_roi)) & \
                            (surface_points_middle[:, 2] > min(z_roi)) & (surface_points_middle[:, 2] < max(z_roi))
            roi_mask_idx = np.where(roi_mask_bool)

        else:
            if x_roi is not None or y_roi is not None or z_roi is not None:
                raise ValueError(f"Either provide X_ROI, Y_ROI, Z_ROI or fn_mask, not both.")

            # read mask from freesurfer mask file
            if not fn_mask.startswith(os.sep):
                fn_mask = os.path.join(mesh_folder, fn_mask)
            if fn_mask.endswith('.mgh') or fn_mask.endswith('.mgz'):
                mask = nib.freesurfer.mghformat.MGHImage.from_filename(fn_mask).dataobj[:]
            else:
                mask = nib.freesurfer.read_morph_data(fn_mask)[:, np.newaxis, np.newaxis]
            roi_mask_idx = np.where(mask > 0.5)

        # redefine connectivity matrix for cropped points (reindexing)
        # get row index where all points are lying inside ROI
        if not (x_roi_wb and y_roi_wb and z_roi_wb):
            con_row_idx = [i for i in range(con.shape[0]) if len(np.intersect1d(con[i, ], roi_mask_idx)) == 3]
            # crop connectivity matrix to ROI
            con_cropped = con[con_row_idx, ]
        else:
            con_cropped = con

        # evaluate new indices of cropped connectivity matrix
        point_idx_before, point_idx_after = np.unique(con_cropped, return_inverse=True)
        con_cropped_reform = np.reshape(point_idx_after, (con_cropped.shape[0], con_cropped.shape[1]))

        # crop points to ROI
        surface_points_middle = surface_points_middle[point_idx_before, ]
        if self.layer == 3:
            surface_points_upper = surface_points_upper[point_idx_before, ]
            surface_points_lower = surface_points_lower[point_idx_before, ]

        # refine
        if refine:
            if not os.path.exists(os.path.join(self.mesh_folder, "roi", "tmp")):
                os.makedirs(os.path.join(self.mesh_folder, "roi", "tmp"))

            mesh = trimesh.Trimesh(vertices=surface_points_middle,
                                   faces=con_cropped_reform)
            roi_fn = os.path.join(self.mesh_folder, "roi", "tmp", "roi.stl")
            mesh.export(roi_fn)

            roi_refined_fn = os.path.join(self.mesh_folder, "roi", "tmp", "roi_refined.stl")
            pynibs.refine_surface(fn_surf=roi_fn,
                                  fn_surf_refined=roi_refined_fn,
                                  center=[0, 0, 0],
                                  radius=np.inf,
                                  verbose=True,
                                  repair=False)

            roi = trimesh.load(roi_refined_fn)
            con_cropped_reform = roi.faces
            surface_points_middle = roi.vertices

            if self.layer == 3:
                roi = []

                for p in [surface_points_upper, surface_points_lower]:
                    mesh = trimesh.Trimesh(vertices=p,
                                           faces=con_cropped_reform)
                    mesh.export(roi_fn)

                    pynibs.refine_surface(fn_surf=roi_fn,
                                          fn_surf_refined=roi_refined_fn,
                                          center=[0, 0, 0],
                                          radius=np.inf,
                                          verbose=True,
                                          repair=False)

                    roi.append(trimesh.load(roi_refined_fn))

                surface_points_upper = roi[0].vertices
                surface_points_lower = roi[1].vertices

        self.node_coord_mid = surface_points_middle
        self.node_number_list = con_cropped_reform

        if self.layer == 3:
            self.node_coord_up = surface_points_upper
            self.node_coord_low = surface_points_lower
            self.tri_center_coord_up = np.average(self.node_coord_up[self.node_number_list], axis=1)
            self.tri_center_coord_low = np.average(self.node_coord_low[self.node_number_list], axis=1)

        self.tri_center_coord_mid = np.average(self.node_coord_mid[self.node_number_list], axis=1)

        if self.layer == 3:
            return surface_points_upper, surface_points_middle, surface_points_lower, con_cropped_reform
        else:
            return surface_points_middle, con_cropped_reform

    def generate_cortical_laminae(self,  # use tuple instead of list for immutability
                                  head_model_mesh,
                                  bbox=None,
                                  laminae=(0.06, 0.4, 0.55, 0.65, 0.85),
                                  layer_ids=("L1", "L23", "L4", "L5", "L6")
                                  ):
        """
        Create the cortical layering with the provided laminar depths.

        Defaults to the standard depths of the laminae in the neo-cortex from layer I to VI
        from "Simulation of transcranial magnetic stimulation in head model with morphologically-realistic
        cortical neurons", Aberra et al., https://doi.org/10.1016/j.brs.2019.10.002

        Parameters
        ----------
        head_model_mesh : simnibs.Msh
            The head model volume mesh.
            Inside the GM compartment of this mesh, the layering will be generated.
        bbox : np.ndarray, optional
            Bounding coordinates of the region of interest.
            Optional, if the mid-layer surface is already existing
            (and can thus be used to determine the bounding coordinates).
        laminae : list of float or tuple of float, default: (0.06, 0.4, 0.55, 0.65, 0.85)
            List of depths of the individual to-be created lamiae.
        layer_ids : typing.List[str], default: ("L1", "L23", "L4", "L5", "L6")
            List of layer identifiers.
        """
        assert bbox is not None or self.node_coord_mid is not None, \
            "Neither mid-layer surface was initialized nor explicit bounding box was provided to construct layer."

        for idx, lam in tqdm.tqdm(enumerate(laminae)):
            self.layers.append(
                CorticalLayer.create_in_roi(
                    layer_id=layer_ids[idx],
                    roi=self,
                    depth=lam,
                    volmesh=head_model_mesh
                )
            )

    def determine_element_idx_in_mesh(self, msh):
        """
        Determines tetrahedra indices of msh where the triangle center points of upper, middle and lower surface
        and the nodes of middle surface are

        Parameters
        ----------
        msh : pynibs.mesh.mesh_struct.TetrahedraLinear
            TetrahedraLinear object.

        Returns
        -------
        RegionOfInterestSurface.tet_idx_tri_center_up : np.ndarray
            (N_points) Tetrahedra indices of TetrahedraLinear object instance where the center points of the
            triangles of the upper surface are.
        RegionOfInterestSurface.tet_idx_tri_center_mid : np.ndarray
            (N_points) Tetrahedra indices of TetrahedraLinear object instance where the center points of the triangles
            of the middle surface are.
        RegionOfInterestSurface.tet_idx_tri_center_low : np.ndarray
            (N_points) Tetrahedra indices of TetrahedraLinear object instance where the center points of the
            triangles of the lower surface are.
        RegionOfInterestSurface.tet_idx_node_coord_mid : np.ndarray
            (N_tri) Tetrahedra indices of TetrahedraLinear object instance where the nodes of the middle
            surface are.
     """
        # determine tetrahedra indices of triangle center points of upper, middle and lower surface
        if self.tri_center_coord_low != [] and self.tri_center_coord_up != []:
            points = [self.tri_center_coord_low,
                      self.tri_center_coord_mid,
                      self.tri_center_coord_up]
        else:
            points = [self.tri_center_coord_mid]

        tet_idx = pynibs.determine_element_idx_in_mesh(fname=None,
                                                       msh=msh,
                                                       points=points,
                                                       compute_baricentric=False)

        if self.tri_center_coord_low != [] and self.tri_center_coord_up != []:
            self.tet_idx_tri_center_low = tet_idx[:, 0]
            self.tet_idx_tri_center_mid = tet_idx[:, 1]
            self.tet_idx_tri_center_up = tet_idx[:, 2]
        else:
            self.tet_idx_tri_center_low = None
            self.tet_idx_tri_center_mid = tet_idx[:, 0]
            self.tet_idx_tri_center_up = None

        # determine tetrahedra indices of nodes of middle surface
        self.tet_idx_node_coord_mid = pynibs.determine_element_idx_in_mesh(fname=None,
                                                                           msh=msh,
                                                                           points=self.node_coord_mid,
                                                                           compute_baricentric=False)

    def decimate(self, fraction=.075):
        """
        Subsample ROI surface based on a decimation factor and return element indices.
        (no Freesurfer surfaces associated with the ROI surface required)

        Parameters
        ----------
        fraction : float, default: .075
            Multiplied by the total number of ROI elements determines
            (approximately) the number of remaining ROI elements after decimation.

        Returns
        -------
        ele_idx : np.ndarray of float
            [approx. fraction * n_ele] Element indices of the subsampled surface; sorted.
        """

        mesh = trimesh.Trimesh(
            vertices=self.node_coord_mid,
            faces=self.node_number_list
        )

        pts, subsample_idcs = trimesh.sample.sample_surface_even(
            mesh=mesh,
            count=int(self.node_number_list.shape[0] * fraction)
        )

        return np.sort(subsample_idcs)

    def subsample(self, dist=10, fn_sphere=None):
        """
        Subsample ROI surface based on a spacing and return element indices
        (Freesurfer surfaces associatd with the ROI surface required)

        Parameters
        ----------
        dist : float
            Distance in mm the subsampled points lie apart.
        fn_sphere : str
            Name of ?.sphere file (freesurfer).

        Returns
        -------
        ele_idx : ndarray of float
            (n_ele) Element indices of the subsampled surface.
        """

        # load sphere surface
        sphere_coords, sphere_faces = nib.freesurfer.io.read_geometry(fn_sphere)

        # sample sphere equally
        r = np.linalg.norm(sphere_coords[0, :])
        a0 = 4 * np.pi * r ** 2

        sphere_points_sampled = pynibs.mesh.utils.sample_sphere(n_points=int(np.ceil(a0 / dist ** 2) // 2 * 2 + 1),
                                                                r=np.linalg.norm(sphere_coords[0, :]))

        # read mask from freesurfer mask file
        mask = nib.freesurfer.mghformat.MGHImage.from_filename(
            os.path.join(self.mesh_folder, self.fn_mask)).dataobj[:].flatten()
        roi_mask_idx = np.where(mask > 0.5)

        con_row_idx = [i for i in range(sphere_faces.shape[0]) if
                       len(np.intersect1d(sphere_faces[i, ], roi_mask_idx)) == 3]

        # crop connectivity matrix to ROI
        con_cropped = sphere_faces[con_row_idx, ]

        p1_tri = sphere_coords[con_cropped[:, 0], :]
        p2_tri = sphere_coords[con_cropped[:, 1], :]
        p3_tri = sphere_coords[con_cropped[:, 2], :]
        tri_center = 1.0 / 3 * (p1_tri + p2_tri + p3_tri)
        ele_idx = []

        for i_p_ss in range(sphere_points_sampled.shape[0]):
            dist = np.linalg.norm(tri_center - sphere_points_sampled[i_p_ss], axis=1)

            if np.min(dist) < 1:
                ele_idx.append(np.argmin(dist))

        return ele_idx


class RegionOfInterestVolume:
    """
    Region of interest (volume) class

    Attributes
    ----------
    node_coord : np.ndarray
        (N_points, 3) Coordinates (x,y,z) of ROI tetrahedra nodes.
    tet_node_number_list : np.ndarray
        (N_tet_roi, 3) Connectivity matrix of ROI tetrahedra.
    tri_node_number_list : np.ndarray
        (N_tri_roi, 3) Connectivity matrix of ROI tetrahedra.
    tet_idx_node_coord : np.ndarray
        (N_points) Tetrahedra indices of TetrahedraLinear object instance where the ROI nodes are.
    tet_idx_tetrahedra_center : np.ndarray
        (N_tet_roi) Tetrahedra indices of TetrahedraLinear object instance where the center points of the ROI
        tetrahedra are.
    tet_idx_triangle_center : np.ndarray
        (N_tri_roi) Tetrahedra indices of TetrahedraLinear object instance where the center points of the ROI triangle
        are. If the ROI is directly generated from the msh instance using "make_roi_volume_from_msh", these
        indices are the triangle indices of the head mesh since the ROI mesh and the head mesh are overlapping. If
        the ROI mesh is not the same as the head mesh, the triangle center of the ROI mesh are always lying in a
        tetrahedra of the head mesh (these indices are given in this case).
    """

    def __init__(self):
        """ Initialize RegionOfInterestVolume class instance """

        self.node_coord = []
        self.tet_node_number_list = []
        self.tri_node_number_list = []
        self.tet_idx_node_coord = []
        self.tet_idx_tetrahedra_center = []
        self.tet_idx_triangle_center = []

    def make_roi_volume_from_msh(self, msh, volume_type='box', x_roi=None, y_roi=None, z_roi=None):
        """
        Generate region of interest (volume) and extract nodes, triangles and tetrahedra from msh instance.

        Parameters
        ----------
        msh: pynibs.mesh.mesh_struct.TetrahedraLinear
            Mesh object instance of type TetrahedraLinear
        volume_type: str
            Type of ROI ('box' or 'sphere')
        x_roi: list of float

            - type = 'box': [Xmin, Xmax] (in mm), whole X range if empty [0,0] or None (left - right)
            - type = 'sphere': origin [x,y,z]
        y_roi: list of float

            - type = 'box': [Ymin, Ymax] (in mm), whole Y range if empty [0,0] or None (anterior - posterior)
            - type = 'sphere': radius (in mm)
        z_roi: list of float

            - type = 'box': [Zmin, Zmax] (in mm), whole Z range if empty [0,0] or None (inferior - superior)
            - type = 'sphere': None

        Returns
        -------
        RegionOfInterestVolume.node_coord : np.ndarray [N_points x 3]
            Coordinates (x,y,z) of ROI tetrahedra nodes
        RegionOfInterestVolume.tet_node_number_list : np.ndarray [N_tet_roi x 3]
            Connectivity matrix of ROI tetrahedra
        RegionOfInterestVolume.tri_node_number_list : np.ndarray [N_tri_roi x 3]
            Connectivity matrix of ROI tetrahedra
        RegionOfInterestVolume.tet_idx_node_coord : np.ndarray [N_points]
            Tetrahedra indices of TetrahedraLinear object instance where the ROI nodes are lying in
        RegionOfInterestVolume.tet_idx_tetrahedra_center : np.ndarray [N_tet_roi]
            Tetrahedra indices of TetrahedraLinear object instance where the center points of the ROI tetrahedra are
            lying in
        RegionOfInterestVolume.tet_idx_triangle_center : np.ndarray [N_tri_roi]
            Tetrahedra indices of TetrahedraLinear object instance where the center points of the ROI triangle are
           . If the ROI is directly generated from the msh instance using "make_roi_volume_from_msh", these
            indices are the triangle indices of the head mesh since the ROI mesh and the head mesh are overlapping. If
            the ROI mesh is not the same as the head mesh, the triangle center of the ROI mesh are always. a
            tetrahedra of the head mesh (these indices are given in this case)
        """

        if volume_type == 'box':
            roi_mask_bool = (msh.points[:, 0] > min(x_roi)) & (msh.points[:, 0] < max(x_roi)) & \
                            (msh.points[:, 1] > min(y_roi)) & (msh.points[:, 1] < max(y_roi)) & \
                            (msh.points[:, 2] > min(z_roi)) & (msh.points[:, 2] < max(z_roi))
            roi_mask_idx = np.where(roi_mask_bool)[0]

        elif volume_type == 'sphere':
            roi_mask_bool = np.linalg.norm(msh.points - np.array(x_roi), axis=1) <= y_roi
            roi_mask_idx = np.where(roi_mask_bool)[0]

        else:
            raise Exception('region of interest type not specified correctly (either box or sphere)')

        self.node_coord = msh.points[roi_mask_idx, :]

        # crop connectivity matrices of tetrahedra and triangles to ROI
        tet_con_row_idx = [i for i in range(msh.tetrahedra.shape[0]) if
                           len(np.intersect1d(msh.tetrahedra[i, ], roi_mask_idx)) == 3]
        tet_con_cropped = msh.tetrahedra[tet_con_row_idx,]

        tri_con_row_idx = [i for i in range(msh.triangles.shape[0]) if
                           len(np.intersect1d(msh.triangles[i, ], roi_mask_idx)) == 3]
        tri_con_cropped = msh.triangles[tri_con_row_idx, ]

        # evaluate new indices of cropped connectivity matrices of tetrahedra and triangles (starts from 0)
        tet_point_idx_before, tet_point_idx_after = np.unique(tet_con_cropped, return_inverse=True)
        self.tet_node_number_list = np.reshape(tet_point_idx_after,
                                               (tet_con_cropped.shape[0], tet_con_cropped.shape[1]))

        tri_point_idx_before, tri_point_idx_after = np.unique(tri_con_cropped, return_inverse=True)
        self.tri_node_number_list = np.reshape(tri_point_idx_after,
                                               (tri_con_cropped.shape[0], tri_con_cropped.shape[1]))

        self.tet_idx_node_coord = None
        self.tet_idx_tetrahedra_center = np.array(tet_con_row_idx)
        self.tet_idx_triangle_center = np.array(tri_con_row_idx)
